from __future__ import annotations

import sys
from abc import abstractmethod
from pathlib import Path
from typing import Any

from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml

from .constants import COGNITE_MODULES, CUSTOM_MODULES, HINT_LEAD_TEXT, MODULES, ROOT_MODULES, URL
from .exceptions import ToolkitFileNotFoundError, ToolkitNotADirectoryError
from .loaders import LOADER_BY_FOLDER_NAME
from .tk_warnings import MediumSeverityWarning
from .utils import find_directory_with_subdirectories

CDF_TOML = CDFToml.load(Path.cwd())


class Hint:
    _indent = " " * 5
    _lead_text = HINT_LEAD_TEXT

    @classmethod
    @abstractmethod
    def _short(cls) -> str: ...

    @classmethod
    @abstractmethod
    def long(cls, *args: Any, **kwargs: Any) -> list[str]: ...

    @classmethod
    def short(cls) -> str:
        return f"{cls._lead_text}{cls._short()}"

    @classmethod
    def _to_hint(cls, lines: list[str]) -> str:
        return cls._lead_text + f"\n{cls._indent}".join(lines)

    @classmethod
    def link(cls, url: str, text: str | None = None) -> str:
        return f"[blue][link={url}]{text or url}[/link][/blue]"


class ModuleDefinition(Hint):
    @classmethod
    def _short(cls) -> str:
        return f"Available resource directories are {sorted(LOADER_BY_FOLDER_NAME)}. {cls.link(URL.configs)} to learn more."

    @classmethod
    def long(cls, missing_modules: set[str | Path] | None = None, organization_dir: Path | None = None) -> str:  # type: ignore[override]
        lines = [
            "A module is a directory with one or more resource directories in it.",
            f"Available resource directories are {sorted(LOADER_BY_FOLDER_NAME)}",
            f"{cls.link(URL.configs)} to learn more",
        ]
        if missing_modules and organization_dir:
            found_directory, subdirectories = find_directory_with_subdirectories(
                next((m for m in missing_modules if isinstance(m, str)), None), organization_dir
            )
            if found_directory:
                lines += [
                    f"For example, the directory {found_directory.as_posix()!r} is not a module, as none of its",
                    f"subdirectories are resource directories. The subdirectories found are: {subdirectories}",
                ]
        return cls._to_hint(lines)


def verify_module_directory(organization_dir: Path, build_env_name: str | None) -> None:
    from .data_classes import BuildConfigYAML

    config_file = BuildConfigYAML.get_filename(build_env_name or "MISSING")

    has_config_yaml = CDF_TOML.cdf.has_user_set_default_env and build_env_name is not None
    if organization_dir != Path.cwd():
        if has_config_yaml:
            content = f"  ┣ {MODULES}/\n  ┗ {config_file}\n"
        else:
            content = f"  ┗ {MODULES}\n"

        panel = Panel(
            f"Toolkit expects the following structure:\n{organization_dir!s}/\n{content}",
            expand=False,
        )
    else:
        if build_env_name:
            content = f"\n  {MODULES}/\n  {config_file}\n"
        else:
            content = f"\n  {MODULES}\n"

        panel = Panel(
            f"Toolkit expects the following structure:{content}",
            expand=False,
        )
    if not organization_dir.is_dir():
        print(panel)
        raise ToolkitNotADirectoryError(f"{organization_dir.as_posix()!r} is not a directory.")

    root_modules = [
        module_dir for root_module in ROOT_MODULES if (module_dir := organization_dir / root_module).exists()
    ]
    if deprecated_modules := [
        root_module for root_module in [CUSTOM_MODULES, COGNITE_MODULES] if (organization_dir / root_module).exists()
    ]:
        if (organization_dir / MODULES).exists():
            suffix = f"Move the modules into {MODULES}/ directory."
        elif len(deprecated_modules) == 1:
            suffix = f"Rename the directory to {MODULES}/."
        else:
            suffix = f"Combine the two into a new {MODULES}/ directory."
        prefix = "Directories" if len(deprecated_modules) > 1 else "Directory"
        verb = "are" if len(deprecated_modules) > 1 else "is"
        MediumSeverityWarning(
            f"{prefix} {', '.join(deprecated_modules)} {verb} deprecated and will be removed in 0.4.0. {suffix}"
        ).print_warning()

    config_path = organization_dir / config_file
    if root_modules and (config_path.is_file() or not has_config_yaml):
        return
    if root_modules or (config_path.is_file() or not has_config_yaml):
        print(panel)
        if not root_modules:
            raise ToolkitNotADirectoryError(
                f"Could not find the {(organization_dir / MODULES).as_posix()!r} directory."
            )
        if not has_config_yaml:
            return
        else:
            raise ToolkitFileNotFoundError(f"Could not find the {config_path.as_posix()!r} file.")

    # Search for the modules directory
    candidate_org = next(
        (
            path
            for path in organization_dir.iterdir()
            if path.is_dir() and any((path / sub).exists() for sub in ROOT_MODULES)
        ),
        None,
    )
    print(panel)
    if candidate_org is not None:
        if candidate_org.is_relative_to(Path.cwd()):
            candidate_rel = candidate_org.relative_to(Path.cwd())
        else:
            candidate_rel = candidate_org
        user_arg = sys.argv
        suggestion = ["cdf"]
        skip_next = False
        found = False
        for arg in user_arg[1:]:
            if arg in ("-o", "--organization-dir"):
                suggestion.append(f"{arg} {candidate_rel}")
                skip_next = True
                found = True
                continue
            if skip_next:
                skip_next = False
                continue
            suggestion.append(arg)
        if not found:
            suggestion.append(f"-o {candidate_rel}")

        print(f"{Hint._lead_text} Did you mean to use the command: '{' '.join(suggestion)}'?")
        cdf_toml = CDFToml.load()
        if not cdf_toml.cdf.has_user_set_default_org:
            print(
                f"{Hint._lead_text} You can specify a 'default_organization_dir = ...' in the 'cdf' section of your "
                f"'{CDFToml.file_name}' file to avoid using the -o/--organization-dir argument"
            )

    raise ToolkitNotADirectoryError(f"Could not find the {(organization_dir / MODULES).as_posix()!r} directory.")



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import difflib
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Sequence
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import capabilities as cap
from cognite.client.data_classes.capabilities import (
    Capability,
    GroupsAcl,
    SecurityCategoriesAcl,
)
from cognite.client.data_classes.iam import (
    Group,
    GroupList,
    GroupWrite,
    GroupWriteList,
    SecurityCategory,
    SecurityCategoryList,
    SecurityCategoryWrite,
    SecurityCategoryWriteList,
)
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console
from rich.markup import escape

from cognite_toolkit._cdf_tk._parameters import ANY_INT, ANY_STR, ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.exceptions import ToolkitWrongResourceError
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.resource_classes import GroupYAML, SecurityCategoriesYAML
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    MediumSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, hash_dict


@dataclass
class _ReplaceMethod:
    """This is a small helper class used in the
    lookup and replace in the ACL scoped ids"""

    lookup_method: Callable[[str, bool], int]
    reverse_lookup_method: Callable[[int], str | None]
    id_name: str


class GroupLoader(ResourceLoader[str, GroupWrite, Group, GroupWriteList, GroupList]):
    folder_name = "auth"
    filename_pattern = r"^(?!.*SecurityCategory$).*"
    kind = "Group"
    resource_cls = Group
    resource_write_cls = GroupWrite
    list_cls = GroupList
    list_write_cls = GroupWriteList
    yaml_cls = GroupYAML
    resource_scopes = frozenset(
        {
            cap.IDScope,
            cap.SpaceIDScope,
            cap.DataSetScope,
            cap.TableScope,
            cap.AssetRootIDScope,
            cap.ExtractionPipelineScope,
            cap.IDScopeLowerCase,
        }
    )
    resource_scope_names = frozenset({scope._scope_name for scope in resource_scopes})  # type: ignore[attr-defined]
    _doc_url = "Groups/operation/createGroups"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None,
        target_scopes: Literal[
            "all_scoped_only",
            "resource_scoped_only",
        ] = "all_scoped_only",
    ):
        super().__init__(client, build_dir, console)
        self.target_scopes = target_scopes

    @property
    def display_name(self) -> str:
        return f"groups({self.target_scopes.removesuffix('_only')})"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[GroupWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
            ]
            if read_only
            else [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
                GroupsAcl.Action.Create,
                GroupsAcl.Action.Delete,
                GroupsAcl.Action.Update,
            ]
        )

        return GroupsAcl(
            actions,
            GroupsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: GroupWrite | Group | dict) -> str:
        if isinstance(item, dict):
            return item["name"]
        return item.name

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"name": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        from .classic_loaders import AssetLoader
        from .data_organization_loaders import DataSetsLoader
        from .datamodel_loaders import SpaceLoader
        from .extraction_pipeline_loaders import ExtractionPipelineLoader
        from .location_loaders import LocationFilterLoader
        from .raw_loaders import RawDatabaseLoader, RawTableLoader
        from .timeseries_loaders import TimeSeriesLoader

        for capability in item.get("capabilities", []):
            for acl, content in capability.items():
                if scope := content.get("scope", {}):
                    if space_ids := scope.get(cap.SpaceIDScope._scope_name, []):
                        if isinstance(space_ids, dict) and "spaceIds" in space_ids:
                            for space_id in space_ids["spaceIds"]:
                                yield SpaceLoader, space_id
                    if data_set_ids := scope.get(cap.DataSetScope._scope_name, []):
                        if isinstance(data_set_ids, dict) and "ids" in data_set_ids:
                            for data_set_id in data_set_ids["ids"]:
                                yield DataSetsLoader, data_set_id
                    if table_ids := scope.get(cap.TableScope._scope_name, []):
                        for db_name, tables in table_ids.get("dbsToTables", {}).items():
                            yield RawDatabaseLoader, RawDatabase(db_name)
                            if isinstance(tables, list):
                                yield from ((RawTableLoader, RawTable(db_name, table)) for table in tables)
                            elif isinstance(tables, dict) and "tables" in tables:
                                for table in tables["tables"]:
                                    yield RawTableLoader, RawTable(db_name, table)
                    if extraction_pipeline_ids := scope.get(cap.ExtractionPipelineScope._scope_name, []):
                        if isinstance(extraction_pipeline_ids, dict) and "ids" in extraction_pipeline_ids:
                            for extraction_pipeline_id in extraction_pipeline_ids["ids"]:
                                yield ExtractionPipelineLoader, extraction_pipeline_id
                    if asset_root_ids := scope.get(cap.AssetRootIDScope._scope_name, []):
                        if isinstance(asset_root_ids, dict) and "rootIds" in asset_root_ids:
                            for asset_root_id in asset_root_ids["rootIds"]:
                                yield AssetLoader, asset_root_id
                    if (ids := scope.get(cap.IDScope._scope_name, [])) or (
                        ids := scope.get(cap.IDScopeLowerCase._scope_name, [])
                    ):
                        loader: type[ResourceLoader] | None = None
                        if acl == cap.DataSetsAcl._capability_name:
                            loader = DataSetsLoader
                        elif acl == cap.ExtractionPipelinesAcl._capability_name:
                            loader = ExtractionPipelineLoader
                        elif acl == cap.TimeSeriesAcl._capability_name:
                            loader = TimeSeriesLoader
                        elif acl == cap.SecurityCategoriesAcl._capability_name:
                            loader = SecurityCategoryLoader
                        elif acl == cap.LocationFiltersAcl._capability_name:
                            loader = LocationFilterLoader
                        if loader is not None and isinstance(ids, dict) and "ids" in ids:
                            for id_ in ids["ids"]:
                                yield loader, id_

    def _substitute_scope_ids(self, group: dict[str, Any], is_dry_run: bool, reverse: bool = False) -> dict[str, Any]:
        replace_method_by_acl = self._create_replace_method_by_acl_and_scope()

        for capability in group.get("capabilities", []):
            for acl, values in capability.items():
                scope = values.get("scope", {})
                if len(scope) != 1:
                    # This will raise an error when the group is loaded.
                    continue
                scope_name, scope_content = next(iter(scope.items()))

                if (acl, scope_name) in replace_method_by_acl:
                    replace_method = replace_method_by_acl[(acl, scope_name)]
                elif scope_name in replace_method_by_acl:
                    replace_method = replace_method_by_acl[scope_name]
                else:
                    continue
                if ids := scope.get(scope_name, {}).get(replace_method.id_name, []):
                    if reverse:
                        cdf_ids = (
                            replace_method.reverse_lookup_method(int_id) if isinstance(int_id, int) else int_id
                            for int_id in ids
                        )
                        values["scope"][scope_name][replace_method.id_name] = [
                            id_ for id_ in cdf_ids if id_ is not None
                        ]
                    else:
                        values["scope"][scope_name][replace_method.id_name] = [
                            replace_method.lookup_method(ext_id, is_dry_run) if isinstance(ext_id, str) else ext_id
                            for ext_id in ids
                        ]
        return group

    def _create_replace_method_by_acl_and_scope(self) -> dict[tuple[str, str] | str, _ReplaceMethod]:
        source = {
            (cap.DataSetsAcl, cap.DataSetsAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            (cap.ExtractionPipelinesAcl, cap.ExtractionPipelinesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            (cap.LocationFiltersAcl, cap.LocationFiltersAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.location_filters.id,
                self.client.lookup.location_filters.external_id,
                id_name="ids",
            ),
            (cap.SecurityCategoriesAcl, cap.SecurityCategoriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.security_categories.id,
                self.client.lookup.security_categories.external_id,
                id_name="ids",
            ),
            (cap.TimeSeriesAcl, cap.TimeSeriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.time_series.id,
                self.client.lookup.time_series.external_id,
                id_name="ids",
            ),
            cap.DataSetScope: _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            cap.ExtractionPipelineScope: _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            cap.AssetRootIDScope: _ReplaceMethod(
                self.client.lookup.assets.id,
                self.client.lookup.assets.external_id,
                id_name="rootIds",
            ),
        }
        # Trick to avoid writing _capability_name and _scope_name for each entry.
        return {
            (key[0]._capability_name, key[1]._scope_name) if isinstance(key, tuple) else key._scope_name: method  # type: ignore[attr-defined]
            for key, method in source.items()
        }

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> GroupWrite:
        is_resource_scoped = any(
            any(scope_name in capability.get(acl, {}).get("scope", {}) for scope_name in self.resource_scope_names)
            for capability in resource.get("capabilities", [])
            for acl in capability
        )

        if self.target_scopes == "all_scoped_only" and is_resource_scoped:
            raise ToolkitWrongResourceError()

        if self.target_scopes == "resource_scoped_only" and not is_resource_scoped:
            raise ToolkitWrongResourceError()

        substituted = self._substitute_scope_ids(resource, is_dry_run)
        try:
            loaded = GroupWrite.load(substituted)
        except ValueError:
            # The GroupWrite class in the SDK will raise a ValueError if the ACI or scope is not valid or unknown.
            loaded = GroupWrite._load(substituted, allow_unknown=True)
            for capability in loaded.capabilities or []:
                if isinstance(capability, cap.UnknownAcl):
                    msg = (
                        f"In group {loaded.name!r}, unknown capability found: {capability.capability_name!r}.\n"
                        "Will proceed with group creation and let the API validate the capability."
                    )
                    if matches := difflib.get_close_matches(capability.capability_name, cap.ALL_CAPABILITIES):
                        msg += f"\nIf the API rejects the capability, could it be that you meant on of: {matches}?"
                    prefix, warning_msg = MediumSeverityWarning(msg).print_prepare()
                    print(prefix, warning_msg)

        return loaded

    def dump_resource(self, resource: Group, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        if not dumped.get("sourceId") and "sourceId" not in local:
            dumped.pop("sourceId", None)
        # RAWAcls are not returned by the API following the spec.
        # If you have a table scoped RAW ACL the spec, and thus user will input
        # tableScope:
        #   dbsToTables
        #    db1:
        #    - tables1
        #    - tables2
        # While the API will return
        # tableScope:
        #   dbsToTables:
        #   db1:
        #     tables: [tables1, tables2]
        # Note the extra keyword 'tables' in the API response.
        for capability in dumped.get("capabilities", []):
            for acl, content in capability.items():
                if acl != cap.RawAcl._capability_name:
                    continue
                if scope := content.get("scope", {}):
                    if table_scope := scope.get(cap.TableScope._scope_name, {}):
                        db_to_tables = table_scope.get("dbsToTables", {})
                        if not db_to_tables:
                            continue
                        for db_name in list(db_to_tables.keys()):
                            tables = db_to_tables[db_name]
                            if isinstance(tables, dict) and "tables" in tables:
                                db_to_tables[db_name] = tables["tables"]
        # When you dump a CDF Group, all the referenced resources should be available in CDF.
        return self._substitute_scope_ids(dumped, is_dry_run=False, reverse=True)

    def create(self, items: Sequence[GroupWrite]) -> GroupList:
        if len(items) == 0:
            return GroupList([])
        return self._create_with_fallback(items, action="create")

    def update(self, items: Sequence[GroupWrite]) -> GroupList:
        # We MUST retrieve all the old groups BEFORE we add the new, if not the new will be deleted
        old_groups = self.client.iam.groups.list(all=True)
        created = self._create_with_fallback(items, action="update")
        created_names = {g.name for g in created}
        to_delete = GroupList([group for group in old_groups if group.name in created_names])
        if to_delete:
            self._delete(to_delete, check_own_principal=False)
        return created

    def _create_with_fallback(self, items: Sequence[GroupWrite], action: Literal["create", "update"]) -> GroupList:
        try:
            return self.client.iam.groups.create(items)
        except CogniteAPIError as e:
            if not (e.code == 400 and "buffer" in e.message.lower() and len(items) > 1):
                raise e
            # Fallback to create one by one
            created_list = GroupList([])
            for item in items:
                try:
                    created = self.client.iam.groups.create(item)
                except CogniteAPIError as e:
                    HighSeverityWarning(f"Failed to {action} group {item.name}. Error: {escape(str(e))}").print_warning(
                        include_timestamp=True, console=self.console
                    )
                else:
                    created_list.append(created)
            return created_list

    def retrieve(self, ids: SequenceNotStr[str]) -> GroupList:
        id_set = set(ids)
        remote = self.client.iam.groups.list(all=True)
        found = [g for g in remote if g.name in id_set]
        return GroupList(found)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        return self._delete(self.retrieve(ids), check_own_principal=True)

    def _delete(self, delete_candidates: GroupList, check_own_principal: bool = True) -> int:
        if check_own_principal:
            print_fun = self.console.print if self.console else print
            try:
                # Let's prevent that we delete groups we belong to
                my_groups = self.client.iam.groups.list()
            except CogniteAPIError as e:
                print_fun(
                    f"[bold red]ERROR:[/] Failed to retrieve the current service principal's groups. Aborting group deletion.\n{e}"
                )
                return 0
            my_source_ids = {g.source_id for g in my_groups if g.source_id}
        else:
            my_source_ids = set()

        to_delete: list[int] = []
        counts_by_name: dict[str, int] = defaultdict(int)
        for group in delete_candidates:
            if group.source_id in my_source_ids:
                HighSeverityWarning(
                    f"Not deleting group {group.name} with sourceId {group.source_id} as it is used by"
                    f"the current service principal. If you want to delete this group, you must do it manually."
                ).print_warning(console=self.console)
            else:
                to_delete.append(group.id)
                counts_by_name[group.name] += 1
        if duplicates := {name for name, count in counts_by_name.items() if count > 1}:
            MediumSeverityWarning(
                f"The following names are used by multiple groups (all will be deleted): {duplicates}"
            ).print_warning(console=self.console)

        failed_deletes = []
        error_str = ""
        try:
            self.client.iam.groups.delete(to_delete)
        except CogniteNotFoundError:
            # Fallback to delete one by one
            for delete_item_id in to_delete:
                try:
                    self.client.iam.groups.delete(delete_item_id)
                except CogniteNotFoundError:
                    # If the group is already deleted, we can ignore the error
                    ...
                except CogniteAPIError as e:
                    error_str = str(e)
                    failed_deletes.append(delete_item_id)
        except CogniteAPIError as e:
            error_str = str(e)
            failed_deletes.extend(to_delete)
        if failed_deletes:
            MediumSeverityWarning(
                f"Failed to delete groups: {humanize_collection(to_delete)}. "
                "These must be deleted manually in the Fusion UI."
                f"Error: {escape(error_str)}"
            ).print_warning(include_timestamp=True, console=self.console)
        return len(to_delete) - len(failed_deletes)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Group]:
        return self.client.iam.groups.list(all=True)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # The Capability class in the SDK class Group implementation is deviating from the API.
        # So we need to modify the spec to match the API.
        for item in spec:
            if item.path[0] == "capabilities" and len(item.path) > 2:
                # Add extra ANY_STR layer
                # The spec class is immutable, so we use this trick to modify it.
                object.__setattr__(item, "path", item.path[:2] + (ANY_STR,) + item.path[2:])
        spec.add(
            ParameterSpec(
                ("capabilities", ANY_INT, ANY_STR), frozenset({"dict"}), is_required=False, _is_nullable=False
            )
        )
        spec.add(
            ParameterSpec(
                ("capabilities", ANY_INT, ANY_STR, "scope", ANYTHING),
                frozenset({"dict"}),
                is_required=True,
                _is_nullable=False,
            )
        )
        return spec

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("capabilities",):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        elif json_path[0] == "capabilities":
            # All sublist inside capabilities are hashable
            return diff_list_hashable(local, cdf)
        elif json_path == ("members",):
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


@final
class GroupAllScopedLoader(GroupLoader):
    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console, "all_scoped_only")

    @property
    def display_name(self) -> str:
        return "all-scoped groups"


@final
class SecurityCategoryLoader(
    ResourceLoader[str, SecurityCategoryWrite, SecurityCategory, SecurityCategoryWriteList, SecurityCategoryList]
):
    filename_pattern = r"^.*SecurityCategory$"  # Matches all yaml files who's stem ends with *SecurityCategory.
    resource_cls = SecurityCategory
    resource_write_cls = SecurityCategoryWrite
    list_cls = SecurityCategoryList
    list_write_cls = SecurityCategoryWriteList
    kind = "SecurityCategory"
    yaml_cls = SecurityCategoriesYAML
    folder_name = "auth"
    dependencies = frozenset({GroupAllScopedLoader})
    _doc_url = "Security-categories/operation/createSecurityCategories"

    @property
    def display_name(self) -> str:
        return "security categories"

    @classmethod
    def get_id(cls, item: SecurityCategoryWrite | SecurityCategory | dict) -> str:
        if isinstance(item, dict):
            return item["name"]
        return cast(str, item.name)

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"name": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SecurityCategoryWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.MemberOf,
            ]
            if read_only
            else [
                SecurityCategoriesAcl.Action.Create,
                SecurityCategoriesAcl.Action.Update,
                SecurityCategoriesAcl.Action.MemberOf,
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.Delete,
            ]
        )

        return SecurityCategoriesAcl(
            actions,
            SecurityCategoriesAcl.Scope.All(),
        )

    def create(self, items: SecurityCategoryWriteList) -> SecurityCategoryList:
        return self.client.iam.security_categories.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SecurityCategoryList:
        names = set(ids)
        categories = self.client.iam.security_categories.list(limit=-1)
        return SecurityCategoryList([c for c in categories if c.name in names])

    def update(self, items: SecurityCategoryWriteList) -> SecurityCategoryList:
        items_by_name = {item.name: item for item in items}
        retrieved = self.retrieve(list(items_by_name.keys()))
        retrieved_by_name = {item.name: item for item in retrieved}
        new_items_by_name = {item.name: item for item in items if item.name not in retrieved_by_name}
        if new_items_by_name:
            created = self.client.iam.security_categories.create(list(new_items_by_name.values()))
            retrieved_by_name.update({item.name: item for item in created})
        return SecurityCategoryList([retrieved_by_name[name] for name in items_by_name])

    def delete(self, ids: SequenceNotStr[str]) -> int:
        retrieved = self.retrieve(ids)
        if retrieved:
            self.client.iam.security_categories.delete([item.id for item in retrieved if item.id])
        return len(retrieved)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[SecurityCategory]:
        return self.client.iam.security_categories.list(limit=-1)



import itertools
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client._cognite_client import CogniteClient
from cognite.client.config import ClientConfig
from cognite.client.data_classes.data_modeling.ids import _load_space_identifier
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.statistics import (
    ProjectStatsAndLimits,
    SpaceInstanceCounts,
    SpaceInstanceCountsList,
)


class StatisticsAPI(APIClient):
    _RESOURCE_PATH = "/models/statistics"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        # This is an alpha API, which requires a specific version.
        super().__init__(config, "v1", cognite_client)
        self._RETRIEVE_LIMIT = 100

    def project(self) -> ProjectStatsAndLimits:
        """`Retrieve project-wide usage data and limits

        Returns the usage data and limits for a project's data modelling usage, including data model schemas
        and graph instances

        Returns:
            ProjectStatsAndLimits: The requested statistics and limits

        """
        response_data = self._get(self._RESOURCE_PATH).json()
        if "project" not in response_data:
            response_data["project"] = self._cognite_client._config.project
        return ProjectStatsAndLimits._load(response_data)

    @overload
    def list(self, space: str) -> SpaceInstanceCounts: ...

    @overload
    def list(self, space: SequenceNotStr[str] | None = None) -> SpaceInstanceCountsList: ...

    def list(self, space: str | SequenceNotStr[str] | None = None) -> SpaceInstanceCounts | SpaceInstanceCountsList:
        """`Retrieve usage data and limits per space

        Args:
            space (str | SequenceNotStr[str] | None): The space or spaces to retrieve statistics for.
                If None, all spaces will be retrieved.

        Returns:
            SpaceInstanceCounts | SpaceInstanceCountsList: InstanceStatsPerSpace if a single space is given, else
                InstanceStatsList (which is a list of InstanceStatsPerSpace)

        """
        if space is None:
            return SpaceInstanceCountsList._load(self._get(self._RESOURCE_PATH + "/spaces").json()["items"])

        is_single = isinstance(space, str)

        ids = _load_space_identifier(space)
        result = SpaceInstanceCountsList._load(
            itertools.chain.from_iterable(
                self._post(self._RESOURCE_PATH + "/spaces/byids", json={"items": chunk.as_dicts()}).json()["items"]
                for chunk in ids.chunked(self._RETRIEVE_LIMIT)
            )
        )
        if is_single:
            return result[0]
        return result



from pathlib import Path

import pytest
from cognite.client.data_classes.data_modeling import DirectRelationReference, NodeId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteTimeSeriesApply

from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeries
from cognite_toolkit._cdf_tk.commands import MigrateTimeseriesCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import (
    ExternalIdMigrationMapping,
    IdMigrationMapping,
    MigrationMappingList,
)


class TestMigrateTimeSeriesCommand:
    @pytest.mark.parametrize(
        "ts, expected",
        [
            (
                ExtendedTimeSeries(
                    external_id="full_ts",
                    data_set_id=123,
                    description="Test description",
                    name="Test Time Series",
                    is_step=True,
                    is_string=False,
                    unit="m/s",
                    unit_external_id="velocity:m-per-sec",
                    pending_instance_id=NodeId("sp_full_ts", "full_ts_id"),
                ),
                CogniteTimeSeriesApply(
                    space="sp_full_ts",
                    external_id="full_ts_id",
                    is_step=True,
                    time_series_type="numeric",
                    name="Test Time Series",
                    description="Test description",
                    source_unit="m/s",
                    unit=DirectRelationReference("cdf_cdm_units", "velocity:m-per-sec"),
                ),
            ),
            (
                ExtendedTimeSeries(
                    external_id="minimum_ts",
                    is_step=False,
                    is_string=True,
                    pending_instance_id=NodeId("sp_step_ts", "step_ts_id"),
                ),
                CogniteTimeSeriesApply(
                    space="sp_step_ts",
                    external_id="step_ts_id",
                    is_step=False,
                    time_series_type="string",
                ),
            ),
        ],
    )
    def test_as_cognite_timeseries(self, ts: ExtendedTimeSeries, expected: CogniteTimeSeriesApply) -> None:
        actual = MigrateTimeseriesCommand.as_cognite_timeseries(ts)

        assert actual.dump() == expected.dump()


class TestMigrationMappingList:
    @pytest.mark.parametrize(
        "content, expected",
        [
            (
                "id,dataSetId,space,externalId\n123,123,sp_full_ts,full_ts_id\n3231,,sp_step_ts,step_ts_id\n",
                MigrationMappingList(
                    [
                        IdMigrationMapping(
                            resource_type="timeseries",
                            id=123,
                            data_set_id=123,
                            instance_id=NodeId("sp_full_ts", "full_ts_id"),
                        ),
                        IdMigrationMapping(
                            resource_type="timeseries",
                            id=3231,
                            data_set_id=None,
                            instance_id=NodeId("sp_step_ts", "step_ts_id"),
                        ),
                    ]
                ),
            ),
            (
                "id,space,externalId\n230,my_space,target_external_id\n",
                MigrationMappingList(
                    [
                        IdMigrationMapping(
                            resource_type="timeseries",
                            id=230,
                            data_set_id=None,
                            instance_id=NodeId("my_space", "target_external_id"),
                        )
                    ]
                ),
            ),
            (
                "externalId,dataSetId,space,externalId\n"
                "full_ts,123,sp_full_ts,full_ts_id\n"
                "minimum_ts,,sp_step_ts,step_ts_id\n",
                MigrationMappingList(
                    [
                        ExternalIdMigrationMapping(
                            resource_type="timeseries",
                            external_id="full_ts",
                            data_set_id=123,
                            instance_id=NodeId("sp_full_ts", "full_ts_id"),
                        ),
                        ExternalIdMigrationMapping(
                            resource_type="timeseries",
                            external_id="minimum_ts",
                            data_set_id=None,
                            instance_id=NodeId("sp_step_ts", "step_ts_id"),
                        ),
                    ]
                ),
            ),
            (
                "externalId,space,externalId\nfull_ts,sp_full_ts,full_ts_id\nminimum_ts,sp_step_ts,step_ts_id\n",
                MigrationMappingList(
                    [
                        ExternalIdMigrationMapping(
                            resource_type="timeseries",
                            external_id="full_ts",
                            data_set_id=None,
                            instance_id=NodeId("sp_full_ts", "full_ts_id"),
                        ),
                        ExternalIdMigrationMapping(
                            resource_type="timeseries",
                            external_id="minimum_ts",
                            data_set_id=None,
                            instance_id=NodeId("sp_step_ts", "step_ts_id"),
                        ),
                    ]
                ),
            ),
        ],
    )
    def test_read_mapping_file(self, content: str, expected: MigrationMappingList, tmp_path: Path) -> None:
        input_file = tmp_path / "mapping_file.csv"
        input_file.write_text(content, encoding="utf-8")
        actual = MigrationMappingList.read_mapping_file(input_file)
        assert actual == expected

    @pytest.mark.parametrize(
        "content, expected_msg",
        [
            pytest.param("", "Mapping file is empty", id="empty_file"),
            pytest.param(
                "space,externalId,id,dataSetId\n",
                (
                    "Invalid mapping file header:\n"
                    " - First column must be 'id' or 'externalId'.\n"
                    " - If there are 4 columns, the second column must be 'dataSetId'.\n"
                    " - Last two columns must be 'space' and 'externalId'."
                ),
                id="invalid header",
            ),
            pytest.param(
                "id,data_set_id,space,externalId\n",
                ("Invalid mapping file header:\n - If there are 4 columns, the second column must be 'dataSetId'."),
                id="invalid header with data_set_id",
            ),
            pytest.param(
                "id,externalId\n",
                "Invalid mapping file header:\n"
                " - Mapping file must have at least 3 columns: id/externalId, space, "
                "externalId.\n"
                " - Last two columns must be 'space' and 'externalId'.",
                id="Too few columns",
            ),
            pytest.param(
                "externalId,dataSetId,space,externalId,myExtra\n",
                "Invalid mapping file header:\n"
                " - Mapping file must have at most 4 columns: id/externalId, dataSetId, "
                "space, externalId.\n"
                " - Last two columns must be 'space' and 'externalId'.",
                id="Too many columns",
            ),
            pytest.param(
                "id,dataSetId,space,externalId\n123,123,sp_full_ts,full_ts_id\ninvalid_id,,sp_step_ts,step_ts_id\n",
                "Invalid ID or dataSetId in row 2: ['invalid_id', '', 'sp_step_ts', "
                "'step_ts_id']. ID and dataSetId must be integers.",
                id="Invalid id value",
            ),
            pytest.param(
                "id,dataSetId,space,externalId\n123,invalid_dataset_id,sp_full_ts,full_ts_id\n",
                "Invalid ID or dataSetId in row 1: ['123', 'invalid_dataset_id', "
                "'sp_full_ts', 'full_ts_id']. ID and dataSetId must be integers.",
                id="Invalid external_id value",
            ),
        ],
    )
    def test_read_invalid_mapping_file(self, content: str, expected_msg: str, tmp_path: Path) -> None:
        input_file = tmp_path / "mapping_file.csv"
        input_file.write_text(content, encoding="utf-8")

        with pytest.raises(ValueError) as exc_info:
            MigrationMappingList.read_mapping_file(input_file)
        assert str(exc_info.value) == expected_msg



from __future__ import annotations

import logging
import time
from typing import Any, Optional

from cognite.client import CogniteClient
from cognite.client.data_classes import Annotation, LabelDefinition

logger = logging.getLogger(__name__)


def call_vision_api_endpoint(
    client: CogniteClient,
    url: str,
    input_data: dict,
    max_get_job_attempts: int = 20,
    headers: Optional[dict[str, Any]] = None,
):
    """Post to an API endpoint, wait for response and return response."""
    # TODO: Use this function in people detector function (https://cognitedata.atlassian.net/browse/DMVP-855)
    res = client.post(url=url, json=input_data, headers=headers).json()

    job_id = res.get("jobId")
    for i in range(max_get_job_attempts):
        logger.info(f"Attempt nr. {i} to get job status.")
        res = client.get(url=f"{url}/{job_id}").json()
        if res.get("status") in ["Queued", "Running"]:
            logger.info(f"API job status: {res.get('status')}")
            time.sleep(10)
        elif res.get("status") == "Completed":
            logger.info("API job completed")
            break
    else:  # Use for/else in case of failed job or timeout
        logger.info(f"API job failed or timed out: {res}")
        return None

    return res


def create_missing_labels(client: CogniteClient, label_ids: list[str]):
    """Create missing labels from a list of label IDs."""
    for label_id in label_ids:
        if not check_label_exists(client, label_id):
            client.labels.create([LabelDefinition(external_id=label_id, name=label_id)])


def check_label_exists(client: CogniteClient, label_id: str):
    """Check if a label exists in CDF given its external_id (label_id)."""
    labels = client.labels.list(external_id_prefix=label_id)
    return any(label.external_id == label_id for label in labels)


def create_annotations(
    client: CogniteClient,
    gauge_reading_result: dict,
    file_id: int,
    gauge_type: str,
    bounding_box_label: str,
    keypoint_label: str,
):
    annotations = gauge_reading_result["items"][0]["predictions"][gauge_type + "GaugePredictions"]
    annotations_list = []

    for annotation in annotations:
        print(f"ANNOTATION TYPE: {annotation}")
        if gauge_type == "digital":
            bounding_box = annotation["boundingBox"]
        else:
            keypoints = annotation["keypointCollection"]["keypoints"]
            bounding_box = annotation["objectDetection"]["boundingBox"]
        # Create bounding box annotation
        bounding_box_annotation = Annotation(
            annotation_type="images.ObjectDetection",
            status="suggested",
            creating_user="cognite-functions",
            creating_app="sdk",
            creating_app_version="4.5.2",
            annotated_resource_type="file",
            annotated_resource_id=file_id,
            data={
                "label": bounding_box_label,
                "boundingBox": bounding_box,
            },
        )
        annotations_list.append(bounding_box_annotation)

        if gauge_type in ["level", "dial"]:
            # Create keypoint annotation
            keypoint_names = list(keypoints.keys())
            keypoint_data = {}
            for i in range(len(keypoint_names)):
                keypoint_name = keypoint_names[i]
                print(keypoint_name)
                point = keypoints[keypoint_name]["point"]
                print(point)
                keypoint_data[keypoint_name] = {"point": point}
                print(keypoint_data)

            keypoint_annotation = Annotation(
                annotation_type="images.KeypointCollection",
                status="suggested",
                creating_user="cognite-functions",
                creating_app="sdk",
                creating_app_version="4.5.2",
                annotated_resource_type="file",
                annotated_resource_id=file_id,
                data={"label": keypoint_label, "keypoints": keypoint_data},
            )
            annotations_list.append(keypoint_annotation)

    client.annotations.create(annotations_list)



import pytest

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.statistics import ProjectStatsAndLimits


@pytest.fixture(scope="session")
def project_usage(toolkit_client: ToolkitClient) -> ProjectStatsAndLimits:
    """Fixture to retrieve project usage statistics."""
    return toolkit_client.data_modeling.statistics.project()


class TestStatisticsAPI:
    def test_list_project_instance_usage(self, project_usage: ProjectStatsAndLimits) -> None:
        assert project_usage.instances.instances < project_usage.instances.instances_limit

    def test_list_space_instance_usage(self, toolkit_client: ToolkitClient) -> None:
        spaces = toolkit_client.data_modeling.spaces.list(limit=1)
        assert len(spaces) > 0
        selected_space = spaces[0].space

        space_usage = toolkit_client.data_modeling.statistics.list(space=selected_space)

        assert space_usage.space == selected_space

    def test_list_all_spaces(self, toolkit_client: ToolkitClient, project_usage: ProjectStatsAndLimits) -> None:
        space_usages = toolkit_client.data_modeling.statistics.list()

        assert len(space_usages) > 0
        total = sum(space_usage.nodes + space_usage.edges for space_usage in space_usages)

        assert total == project_usage.instances.instances



from __future__ import annotations

import shutil
from collections import Counter
from importlib import resources
from pathlib import Path
from typing import Any, Literal, Optional

import questionary
import typer
from packaging.version import Version
from packaging.version import parse as parse_version
from rich import print
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from rich.progress import track
from rich.rule import Rule
from rich.table import Table
from rich.tree import Tree

import cognite_toolkit
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import _cli_commands as CLICommands
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands._changes import (
    UPDATE_IMAGE_VERSION_DOCSTRING,
    UPDATE_MODULE_VERSION_DOCSTRING,
    AutomaticChange,
    Changes,
    ManualChange,
    UpdateDockerImageVersion,
    UpdateModuleVersion,
)
from cognite_toolkit._cdf_tk.constants import (
    BUILTIN_MODULES,
    MODULES,
    SUPPORT_MODULE_UPGRADE_FROM_VERSION,
    EnvType,
)
from cognite_toolkit._cdf_tk.data_classes import (
    BuildConfigYAML,
    Environment,
    InitConfigYAML,
    ModuleLocation,
    ModuleResources,
    Package,
    Packages,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitRequiredValueError, ToolkitValueError
from cognite_toolkit._cdf_tk.hints import verify_module_directory
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection, read_yaml_file
from cognite_toolkit._cdf_tk.utils.file import safe_read, safe_rmtree, safe_write, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.modules import module_directory_from_path
from cognite_toolkit._cdf_tk.utils.repository import FileDownloader
from cognite_toolkit._version import __version__

custom_style_fancy = questionary.Style(
    [
        ("qmark", "fg:#673ab7"),  # token in front of the question
        ("question", "bold"),  # question text
        ("answer", "fg:#f44336 bold"),  # submitted answer text behind the question
        ("pointer", "fg:#673ab7 bold"),  # pointer used in select and checkbox prompts
        ("highlighted", "fg:#673ab7 bold"),  # pointed-at choice in select and checkbox prompts
        ("selected", "fg:#673ab7"),  # style for a selected item of a checkbox
        ("separator", "fg:#cc5454"),  # separator in lists
        ("instruction", ""),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        ("disabled", "fg:#858585 italic"),  # disabled choices for select and checkbox prompts
    ]
)

INDENT = "  "
POINTER = INDENT + "▶"


_FILE_DOWNLOADERS_BY_TYPE: dict[str, type[FileDownloader]] = {
    downloader_cls._type: downloader_cls  # type: ignore
    for downloader_cls in FileDownloader.__subclasses__()
}


class ModulesCommand(ToolkitCommand):
    def __init__(self, print_warning: bool = True, skip_tracking: bool = False, silent: bool = False):
        super().__init__(print_warning, skip_tracking, silent)
        self._builtin_modules_path = Path(resources.files(cognite_toolkit.__name__)) / BUILTIN_MODULES  # type: ignore [arg-type]

    @classmethod
    def _create_tree(cls, item: Packages) -> Tree:
        module_locations = [module for package in item.values() for module in package.modules]
        data: dict[str, Any] = {}
        for module in module_locations:
            parts = module.relative_path.parts
            current = data
            for part in parts:
                if part not in current:
                    current[part] = {}
                current = current[part]

        return cls._build_tree(data, Tree(MODULES))

    @classmethod
    def _build_tree(cls, data: dict[str, Any], tree: Tree) -> Tree:
        for key, value in data.items():
            subtree = tree.add(key)
            if isinstance(value, dict):
                cls._build_tree(value, subtree)
        return tree

    def _create(
        self,
        organization_dir: Path,
        selected_packages: Packages,
        environments: list[EnvType],
        mode: Literal["new", "clean", "update"] | None,
        download_data: bool = False,
    ) -> None:
        modules_root_dir = organization_dir / MODULES
        if mode == "clean" and modules_root_dir.is_dir():
            print(f"{INDENT}[yellow]Clearing directory[/]")
            safe_rmtree(modules_root_dir)

        modules_root_dir.mkdir(parents=True, exist_ok=True)

        seen_modules: set[Path] = set()
        selected_paths: set[Path] = set()
        downloader_by_repo: dict[str, FileDownloader] = {}

        extra_resources: set[Path] = set()
        for package_name, package in selected_packages.items():
            print(f"{INDENT}[{'yellow' if mode == 'clean' else 'green'}]Creating {package_name}[/]")

            for module in package.modules:
                if module.dir in seen_modules:
                    # A module can be part of multiple packages
                    continue
                seen_modules.add(module.dir)
                # Add the module and its parent paths to the selected paths, use to load the default.config.yaml
                # files
                selected_paths.update(module.parent_relative_paths)
                selected_paths.add(module.relative_path)
                if module.definition:
                    extra_resources.update(module.definition.extra_resources)

                print(f"{INDENT * 2}[{'yellow' if mode == 'clean' else 'green'}]Creating module {module.name}[/]")
                target_dir = modules_root_dir / module.relative_path
                if Path(target_dir).exists() and mode == "update":
                    if questionary.confirm(
                        f"{INDENT}Module {module.name} already exists in folder {target_dir}. Would you like to overwrite?",
                        default=False,
                    ).ask():
                        safe_rmtree(target_dir)
                    else:
                        continue
                ignore_patterns = ["default.*"]
                if package.name == "quickstart" and module.name != "cdf_ingestion":
                    ignore_patterns.extend(["workflows", "auth"])

                shutil.copytree(module.dir, target_dir, ignore=shutil.ignore_patterns(*ignore_patterns))

                if module.definition is not None and download_data:
                    for example_data in module.definition.data:
                        if example_data.repo not in downloader_by_repo:
                            try:
                                downloader_cls = _FILE_DOWNLOADERS_BY_TYPE[example_data.repo_type]
                            except KeyError:
                                self.warn(
                                    MediumSeverityWarning(
                                        f"Unsupported repo type for example data: {example_data.repo_type}"
                                    )
                                )
                                continue
                            downloader_by_repo[example_data.repo] = downloader_cls(example_data.repo)

                        downloader = downloader_by_repo[example_data.repo]
                        downloader.copy(example_data.source, target_dir / example_data.destination)

        if extra_resources:
            created_by_module: dict[Path, int] = Counter()
            for extra in extra_resources:
                module_dir = module_directory_from_path(extra)
                extra_full_path = self._builtin_modules_path / extra
                target_path = modules_root_dir / extra
                if target_path.exists():
                    # Assume that the user has already created this shared resource
                    continue
                if extra_full_path.is_file():
                    target_path.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(extra_full_path, target_path)
                elif extra_full_path.is_dir():
                    target_path.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copytree(extra_full_path, target_path)
                else:
                    print(f"{INDENT}[red]Extra resource {extra_full_path} not found[/].")
                    continue
                created_by_module[module_dir] += 1
                selected_paths.add(module_dir)
                selected_paths.update(module_dir.parents)

            for module_dir, count in created_by_module.items():
                if count > 0:
                    print(f"{INDENT}Created {count} shared resources in {module_dir.as_posix()!r}.")

        for environment in environments:
            if mode == "update":
                config_init = InitConfigYAML.load_existing(
                    safe_read(Path(organization_dir) / f"config.{environment}.yaml"), environment
                ).load_defaults(self._builtin_modules_path, selected_paths)
            else:
                ignore_variable_patterns: list[tuple[str, ...]] | None = None
                if len(selected_packages) == 1 and "quickstart" in selected_packages:
                    ignore_variable_patterns = []

                    for to_ignore in ["sourcesystem", "contextualization"]:
                        ignore_variable_patterns.extend(
                            [
                                ("modules", to_ignore, "*", "workflow"),
                                ("modules", to_ignore, "*", "workflowClientId"),
                                ("modules", to_ignore, "*", "workflowClientSecret"),
                                ("modules", to_ignore, "*", "groupSourceId"),
                            ]
                        )

                config_init = InitConfigYAML(
                    Environment(
                        name=environment,
                        project=f"<my-project-{environment}>",
                        validation_type=environment,
                        selected=[f"{MODULES}/"],
                    )
                ).load_defaults(self._builtin_modules_path, selected_paths, ignore_variable_patterns)

                if len(selected_packages) == 1 and "quickstart" in selected_packages:
                    config_init.lift()

            print(
                f"{INDENT}[{'yellow' if mode == 'clean' else 'green'}]{'Updating' if mode == 'update' else 'Creating'} config.{environment}.yaml[/]"
            )
            safe_write(Path(organization_dir) / f"config.{environment}.yaml", config_init.dump_yaml_with_comments())

        cdf_toml_content = self.create_cdf_toml(organization_dir, environments[0] if environments else "dev")

        destination = Path.cwd() / CDFToml.file_name
        if destination.exists():
            print(f"{INDENT}[yellow]cdf.toml file already exists. Skipping creation.")
        else:
            destination.write_text(cdf_toml_content, encoding="utf-8")

    def create_cdf_toml(self, organization_dir: Path, env: EnvType = "dev") -> str:
        cdf_toml_content = safe_read(self._builtin_modules_path / CDFToml.file_name)
        if organization_dir != Path.cwd():
            cdf_toml_content = cdf_toml_content.replace(
                "#<PLACEHOLDER>",
                f'''
default_organization_dir = "{organization_dir.name}"''',
            )
        else:
            cdf_toml_content = cdf_toml_content.replace("#<PLACEHOLDER>", "")
        cdf_toml_content = cdf_toml_content.replace("<DEFAULT_ENV_PLACEHOLDER>", env)
        return cdf_toml_content

    def init(
        self,
        organization_dir: Optional[Path] = None,
        select_all: bool = False,
        clean: bool = False,
        user_select: str | None = None,
        user_environments: list[str] | None = None,
        user_download_data: bool | None = None,
    ) -> None:
        if not organization_dir:
            new_line = "\n    "
            message = (
                f"Which directory would you like to create templates in? (default: current directory){new_line}"
                f"HINT Use an organization directory if you use the repository for more than Toolkit. "
                f"If not, use the current (repository root) directory '.':"
            )
            organization_dir_raw = questionary.text(message=message, default="").ask()
            organization_dir = Path(organization_dir_raw.strip())

        modules_root_dir = organization_dir / MODULES
        packages = Packages().load(self._builtin_modules_path)

        if select_all:
            print(Panel("Instantiating all available modules"))
            mode = self._verify_clean(modules_root_dir, clean)
            self._create(
                organization_dir=organization_dir, selected_packages=packages, environments=["dev", "prod"], mode=mode
            )
            return
        is_interactive = user_select is not None
        if not is_interactive:
            print("\n")
            print(
                Panel(
                    "\n".join(
                        [
                            "Wizard for selecting initial modules"
                            "The modules are thematically bundled in packages you can choose between. You can add more by repeating the process.",
                            "You can use the arrow keys ⬆ ⬇  on your keyboard to select modules, and press enter ⮐  to continue with your selection.",
                        ]
                    ),
                    title="Select initial modules",
                    style="green",
                    padding=(1, 2),
                )
            )
        mode = self._verify_clean(modules_root_dir, clean)

        print(f"  [{'yellow' if mode == 'clean' else 'green'}]Using directory [bold]{organization_dir}[/]")

        if user_select is None:
            selected = self._select_packages(packages)
        else:
            selected = Packages([v for k, v in packages.items() if k == user_select])
            if not selected:
                raise ToolkitValueError(f"Package {user_select} not found.")

        if "bootcamp" in selected:
            bootcamp_org = Path.cwd() / "ice-cream-dataops"
            if bootcamp_org != organization_dir:
                # Need to rerun the mode as bootcamp has a hardcoded organization directory
                mode = self._verify_clean(bootcamp_org / MODULES, clean)
            # We only ask to verify if the user has not already selected overwrite in the _verify_clean method
            if mode == "clean" or questionary.confirm("Would you like to continue with creation?", default=True).ask():
                self._create(bootcamp_org, selected, ["test"], mode)
            raise typer.Exit()

        if (
            not is_interactive
            and not questionary.confirm("Would you like to continue with creation?", default=True).ask()
        ):
            print("Exiting...")
            raise typer.Exit()

        if user_environments is None:
            environments = questionary.checkbox(
                "Which environments would you like to include?",
                instruction="Use arrow up/down, press space to select item(s) and enter to save",
                choices=[
                    questionary.Choice(title="dev", checked=True),
                    questionary.Choice(title="prod", checked=True),
                    questionary.Choice(title="staging", checked=False),
                ],
                qmark=INDENT,
                pointer=POINTER,
                style=custom_style_fancy,
            ).ask()
        else:
            environments = user_environments

        if user_download_data is None:
            download_data = self._get_download_data(selected)
        else:
            download_data = user_download_data
        self._create(organization_dir, selected, environments, mode, download_data)

        print(
            Panel(
                f"Modules have been prepared in [bold]{organization_dir}[/].",
                style="green",
            )
        )

        if "custom" in selected:
            print(
                Panel(
                    "Please check out https://docs.cognite.com/cdf/deploy/cdf_toolkit/guides/modules/custom for guidance on writing custom modules",
                )
            )
        if not is_interactive:
            raise typer.Exit()

    @staticmethod
    def _get_download_data(selected: Packages) -> bool:
        example_data = {
            module.name for package in selected.values() for module in package.modules if module.has_example_data
        }
        download_data = False
        if example_data:
            download_data = questionary.confirm(
                f"The modules {humanize_collection(example_data)} has example data. Would you like to download it?",
                default=True,
            ).ask()
        return download_data

    def _select_modules_in_package(self, package: Package) -> list[ModuleLocation]:
        dependencies: set[str] = set()
        for module in package.modules:
            for dependency in module.dependencies:
                dependencies.add(dependency)

        choices = sorted(
            [
                questionary.Choice(
                    title=module.title,
                    value=module,
                    checked=module.name in dependencies or module.is_selected_by_default,
                    disabled="required" if module.name in dependencies else None,
                )
                for module in package.modules
            ],
            key=lambda choice: not choice.checked,
        )

        return questionary.checkbox(
            f"Which modules in {package.name} would you like to add?",
            instruction="Use arrow up/down, press space to select item(s) and enter to save",
            choices=choices,
            qmark=INDENT,
            pointer=POINTER,
            style=custom_style_fancy,
        ).ask()

    def _select_packages(self, packages: Packages, existing_module_names: list[str] | None = None) -> Packages:
        adding_to_existing = False
        if existing_module_names is not None:
            adding_to_existing = True
            for item in packages.values():
                item.modules = [module for module in item.modules if module.name not in existing_module_names]

        selected = Packages()

        while True:
            if "bootcamp" in selected:
                break
            if len(selected) > 0:
                print("\n[bold]You have selected the following:[/]\n")

                tree = self._create_tree(selected)
                print(Padding.indent(tree, 5))
                print("\n")

                if not questionary.confirm("Would you like to make changes to the selection?", default=False).ask():
                    break

            if not any([len(package.modules) > 0 for package in packages.values()]):
                print("No more modules available.")
                break

            choices = [
                questionary.Choice(
                    title=f"{package.title}: {package.description} ({len(package.modules)})", value=package
                )
                for package in [
                    package for package in packages.values() if len(package.modules) > 0 or package.name == "empty"
                ]
            ]

            package: Package = questionary.select(
                "Which package would you like to use?",
                instruction="Use arrow up/down and ⮐  to save",
                choices=choices,
                pointer=POINTER,
                style=custom_style_fancy,
            ).ask()

            if package is None:
                raise typer.Exit(code=0)

            if package.can_cherry_pick and (
                len(package.modules) > 1 or (adding_to_existing and len(package.modules) > 0)
            ):
                selection = self._select_modules_in_package(package)
            else:
                selection = package.modules

            selected[package.name] = Package(
                name=package.name,
                title=package.title,
                description=package.description,
                modules=selection,
            )
        return selected

    @staticmethod
    def _verify_clean(modules_root_dir: Path, clean: bool) -> Literal["new", "clean"]:
        if clean:
            return "clean"
        if not modules_root_dir.is_dir():
            return "new"
        user_selection = questionary.select(
            f"Directory {modules_root_dir} already exists. What would you like to do?",
            choices=[
                questionary.Choice("Abort", "abort"),
                questionary.Choice("Overwrite (clean existing)", "clean"),
            ],
            pointer=POINTER,
            style=custom_style_fancy,
            instruction="use arrow up/down and " + "⮐ " + " to save",
        ).ask()
        if user_selection == "abort":
            print("Aborting...")
            raise typer.Exit()
        return "clean"

    def upgrade(self, organization_dir: Path, verbose: bool = False) -> Changes:
        module_version = self._get_module_version(organization_dir)
        cli_version = parse_version(__version__)

        if cli_version < module_version:
            upgrade = "poetry add cognite-toolkit@" if CLICommands.use_poetry() else "pip install cognite-toolkit=="
            print(
                f"Modules are at a higher version ({module_version}) than the installed CLI ({__version__})."
                f"Please upgrade the CLI to match the modules: `{upgrade}{module_version}`."
            )
            return Changes()

        if cli_version == module_version:
            print("The modules are already at the same version as the CLI.")
            return Changes()

        if module_version < Version(SUPPORT_MODULE_UPGRADE_FROM_VERSION):
            print(
                f"The modules upgrade command is not supported for versions below {SUPPORT_MODULE_UPGRADE_FROM_VERSION}."
            )
            return Changes()

        if not CLICommands.use_git():
            self.warn(MediumSeverityWarning("git is not installed. It is strongly recommended to use version control."))
        else:
            if not CLICommands.has_initiated_repo():
                self.warn(MediumSeverityWarning("git repository not initiated. Did you forget to run `git init`?"))
            else:
                if CLICommands.has_uncommitted_changes():
                    print("Uncommitted changes detected. Please commit your changes before upgrading the modules.")
                    return Changes()
        # Update the docstring of the change 'UpdateModuleVersion' to be more informative
        UpdateModuleVersion.__doc__ = UPDATE_MODULE_VERSION_DOCSTRING.format(
            module_version=module_version, cli_version=cli_version
        )
        UpdateDockerImageVersion.__doc__ = UPDATE_IMAGE_VERSION_DOCSTRING.format(
            module_version=module_version, cli_version=cli_version
        )

        workflow_dir = Path.cwd() / ".github" / "workflows"
        changes = Changes.load(module_version, organization_dir, workflow_dir if workflow_dir.exists() else None)
        if not changes:
            print("No changes required.")
            return changes

        print(
            Panel(
                f"Found {len(changes)} changes from {module_version} to {cli_version}",
                title="Upgrade Modules",
                style="green",
                expand=False,
            )
        )

        total_changed: set[Path] = set()
        iterable = changes if verbose else track(changes, description="Applying changes")
        for change in iterable:
            color = "green"
            changed_files: set[Path] = set()
            if change.has_file_changes:
                if isinstance(change, AutomaticChange):
                    changed_files = change.do()
                    color = "yellow" if changed_files else "green"
                    total_changed.update([file.resolve(strict=False) for file in changed_files])
                elif isinstance(change, ManualChange):
                    changed_files = change.needs_to_change()
                    color = "red" if changed_files else "green"
            if verbose:
                print(
                    Panel(
                        f"Change: {type(change).__name__}",
                        style=color,
                        expand=False,
                    )
                )
                if not changed_files and change.has_file_changes:
                    suffix = "have been changed" if isinstance(change, AutomaticChange) else "need to be changed"
                    print(f"No files {suffix}.")
            else:
                if isinstance(change, ManualChange):
                    print(Markdown(change.instructions(changed_files)))
                elif isinstance(change, AutomaticChange) and verbose:
                    print("The following files have been changed:")
                    for file in changed_files:
                        if file.is_relative_to(organization_dir):
                            print(Markdown(f"  - {file.relative_to(organization_dir).as_posix()}"))
                        else:
                            print(Markdown(f"  - {file.as_posix()}"))
            if verbose:
                print(Markdown(change.__doc__ or "Missing description."))
                print(Rule())

        use_git = CLICommands.use_git() and CLICommands.has_initiated_repo()
        summary = ["All automatic changes have been applied."]
        color = "green"
        if total_changed:
            summary.append(f"A total of {len(total_changed)} files have been changed.")
        else:
            summary.append("No files have been changed.")
        if manual_changes := [change for change in changes if isinstance(change, ManualChange)]:
            summary.append(
                f"A total of {len(manual_changes)} changes require manual intervention: {', '.join([type(change).__name__ for change in manual_changes])}."
            )
            color = "yellow"
        if use_git and total_changed:
            summary.append("\nPlease review the changes and commit them if you are satisfied.")
            summary.append("You can use `git diff` to see the changes or use your IDE to inspect the changes.")
            summary.append(
                "If you are not satisfied with the changes, you can use `git checkout -- <file>` to revert "
                "a file or `git checkout .` to revert all changes."
            )
        print(Panel("\n".join(summary), title="Upgrade Complete", style=color, expand=False))
        return changes

    @staticmethod
    def _get_module_version(project_path: Path) -> Version:
        cdf_toml = CDFToml.load(use_singleton=False)
        # After `0.3.0` the version is in the CDF TOML file
        if cdf_toml.is_loaded_from_file and cdf_toml.modules.version:
            return parse_version(cdf_toml.modules.version)

        elif (system_yaml := project_path / "_system.yaml").exists():
            # From 0.2.0a3 we have the _system.yaml on the root of the project
            content = read_yaml_file(system_yaml)
        elif (system_yaml := project_path / "cognite_modules" / "_system.yaml").exists():
            # Up to 0.2.0a2 we have the _system.yaml in the cognite_modules folder
            content = read_yaml_file(system_yaml)
        else:
            raise ToolkitRequiredValueError(
                "No 'cdf.toml' or '_system.yaml' file found in project."
                "This is needed to determine the version of the modules."
            )
        return parse_version(content.get("cdf_toolkit_version", "0.0.0"))

    def list(self, organization_dir: Path, build_env_name: str | None) -> None:
        if organization_dir in {Path("."), Path("./")}:
            organization_dir = Path.cwd()
        verify_module_directory(organization_dir, build_env_name)
        modules = ModuleResources(organization_dir, build_env_name)

        table = Table(title=f"{build_env_name} {organization_dir.name} modules")
        table.add_column("Module Name", style="bold")
        table.add_column("Resource Folders", style="bold")
        table.add_column("Resources", style="bold")
        table.add_column("Build Warnings", style="bold")
        table.add_column("Build Result", style="bold")
        table.add_column("Location", style="bold")

        for module in modules.list():
            if module.status == "Success":
                status = f"[green]{module.status}[/]"
            else:
                status = f"[red]{module.status}[/]"
            if module.warning_count > 0:
                warning_count = f"[yellow]{module.warning_count:,}[/]"
            else:
                warning_count = f"{module.warning_count:,}"

            table.add_row(
                module.name,
                f"{len(module.resources):,}",
                f"{sum(len(resources) for resources in module.resources.values()):,}",
                warning_count,
                status,
                module.location.path.as_posix(),
            )

        print(table)

    def add(self, organization_dir: Path) -> None:
        verify_module_directory(organization_dir, None)
        environments = [env for env in EnvType.__args__ if (organization_dir / f"config.{env}.yaml").exists()]  # type: ignore[attr-defined]
        if environments:
            build_env = environments[0]
        else:
            default = BuildConfigYAML.load_default(organization_dir)
            print(
                f"No config.[env].yaml files found in {organization_dir}."
                f"This is required to add modules. Creating {default.filename}."
            )
            safe_write(default.filepath, yaml_safe_dump(default.dump()))
            environments.append(default.environment.validation_type)
            build_env = default.environment.validation_type

        existing_module_names = [module.name for module in ModuleResources(organization_dir, build_env).list()]
        available_packages = Packages().load(self._builtin_modules_path)

        added_packages = self._select_packages(available_packages, existing_module_names)

        download_data = self._get_download_data(added_packages)
        self._create(organization_dir, added_packages, environments, "update", download_data)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Hashable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar

from rich.markup import escape

from cognite_toolkit._cdf_tk.constants import HINT_LEAD_TEXT

from .base import SeverityLevel, ToolkitWarning


@dataclass(frozen=True)
class FileReadWarning(ToolkitWarning, ABC):
    severity: ClassVar[SeverityLevel]
    filepath: Path

    def group_key(self) -> tuple[Any, ...]:
        return (self.filepath,)

    def group_header(self) -> str:
        return f"    In File {self.filepath.as_posix()!r}"

    def __str__(self) -> str:
        return self.get_message()


@dataclass(frozen=True)
class FileExistsWarning(FileReadWarning):
    severity = SeverityLevel.MEDIUM
    extra_message: str = ""

    def get_message(self) -> str:
        msg = f"{type(self).__name__}: The file {self.filepath} already exists."
        if self.extra_message:
            msg += f" {self.extra_message}"
        return msg


@dataclass(frozen=True)
class IdentifiedResourceFileReadWarning(FileReadWarning, ABC):
    id_value: str
    id_name: str


@dataclass(frozen=True)
class YAMLFileWarning(FileReadWarning, ABC):
    def __post_init__(self) -> None:
        if self.filepath.suffix not in {".yaml", ".yml"}:
            raise ValueError(f"Expected a YAML file, got {self.filepath.suffix}.")


@dataclass(frozen=True)
class YAMLFileWithElementWarning(YAMLFileWarning, ABC):
    # None is a dictionary, number is a list
    element_no: int | None
    path: tuple[str | int, ...]

    @property
    def _location(self) -> str:
        if self.element_no is not None:
            value = f" in entry {self.element_no} "
        else:
            value = ""
        if len(self.path) <= 1:
            return f"{value}"
        else:
            return f"{value} in section {self.path!r}"


@dataclass(frozen=True)
class DuplicatedItemWarning(YAMLFileWarning):
    severity = SeverityLevel.MEDIUM
    identifier: Hashable
    first_location: Path

    def get_message(self) -> str:
        return (
            f"{type(self).__name__}: Duplicated item with identifier "
            f"{self.identifier!r} first seen in {self.first_location.name!r}."
        )


@dataclass(frozen=True)
class UnknownResourceTypeWarning(YAMLFileWarning):
    severity = SeverityLevel.MEDIUM

    suggestion: str | None

    def get_message(self) -> str:
        msg = f"{type(self).__name__}: File {self.filepath.as_posix()!r}."
        if self.suggestion:
            msg += f"\n[blue]HINT: [/blue]{self.suggestion}"
        return msg


@dataclass(frozen=True)
class UnusedParameterWarning(YAMLFileWithElementWarning):
    severity = SeverityLevel.LOW
    actual: str

    def get_message(self) -> str:
        message = f"{type(self).__name__}: Parameter {self.actual!r} is not used{self._location}."
        if self.actual == "dataSetId" or self.actual == "data_set_id":
            message += f" {HINT_LEAD_TEXT}The parameter 'dataSetExternalId' should be used instead."
        return message


@dataclass(frozen=True)
class UnresolvedVariableWarning(FileReadWarning):
    severity = SeverityLevel.HIGH
    variable: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Variable {self.variable!r} is not resolved in {self.filepath}."


@dataclass(frozen=True)
class ResourceMissingIdentifierWarning(YAMLFileWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    message: ClassVar[str] = "The resource is missing an identifier:"
    resource: str
    ext_id_type: str

    def get_message(self) -> str:
        message = f"The {self.resource} {self.filepath} is missing the {self.ext_id_type} field(s)."
        return message


@dataclass(frozen=True)
class NamingConventionWarning(YAMLFileWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: ClassVar[str] = "The naming convention is not followed:"
    resource: str
    ext_id_type: str
    external_id: str

    @property
    @abstractmethod
    def recommendation(self) -> str:
        raise NotImplementedError()

    def get_message(self) -> str:
        message = (
            f"The {self.ext_id_type} identifier [bold]{self.external_id!r}[/bold] of the resource {self.resource} "
            f"does not follow the recommended naming convention {self.recommendation}"
        )
        return message


@dataclass(frozen=True)
class PrefixConventionWarning(NamingConventionWarning):
    prefix: str

    @property
    def recommendation(self) -> str:
        return f"of prefixing with {self.prefix!r}."


@dataclass(frozen=True)
class NamespacingConventionWarning(NamingConventionWarning):
    namespace: str

    @property
    def recommendation(self) -> str:
        return f"of using {self.namespace!r} as separator."


@dataclass(frozen=True)
class CaseTypoWarning(UnusedParameterWarning):
    expected: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Got {self.actual!r}. Did you mean {self.expected!r}?{self._location}."


@dataclass(frozen=True)
class MissingRequiredParameterWarning(YAMLFileWithElementWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    expected: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Missing required parameter {self.expected!r}{self._location}."


@dataclass(frozen=True)
class MissingRequiredIdentifierWarning(YAMLFileWithElementWarning):
    expected: tuple[str, ...]

    def get_message(self) -> str:
        return f"{type(self).__name__}: Missing required identifier {self.expected!r}{self._location}."


@dataclass(frozen=True)
class MissingReferencedWarning(YAMLFileWithElementWarning):
    message: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: {self.message}."


@dataclass(frozen=True)
class TemplateVariableWarning(IdentifiedResourceFileReadWarning):
    path: str

    def get_message(self) -> str:
        return f"{type(self).__name__}: Variable {self.id_name!r} has value {self.id_value!r} in file: {self.filepath.name}. Did you forget to change it?"


@dataclass(frozen=True)
class DataSetMissingWarning(IdentifiedResourceFileReadWarning):
    severity = SeverityLevel.MEDIUM
    resource_name: str

    def get_message(self) -> str:
        # Avoid circular import
        from cognite_toolkit._cdf_tk.loaders import TransformationLoader

        if self.filepath.parent.name == TransformationLoader.folder_name:
            return f"{type(self).__name__}: It is recommended to use a data set if source or destination can be scoped with a data set. If not, ignore this warning."
        else:
            return f"{type(self).__name__}: It is recommended that you set dataSetExternalId for {self.resource_name}. This is missing in {self.filepath.name}. Did you forget to add it?"


@dataclass(frozen=True)
class SourceFileModifiedWarning(FileReadWarning):
    severity = SeverityLevel.ERROR

    def get_message(self) -> str:
        message = (
            f"{type(self).__name__}: The source file {self.filepath} has been modified since the last build. "
            "Please rebuild the project."
        )
        return message


@dataclass(frozen=True)
class MissingFileWarning(FileReadWarning):
    severity = SeverityLevel.MEDIUM
    attempted_check: str

    def get_message(self) -> str:
        message = f"{type(self).__name__}: The file {self.filepath} is missing. Cannot verify {self.attempted_check}."
        return message


@dataclass(frozen=True)
class EnvironmentVariableMissingWarning(FileReadWarning):
    severity = SeverityLevel.HIGH
    variables: frozenset[str]
    identifiers: frozenset[Hashable] | None = None

    def get_message(self) -> str:
        from cognite_toolkit._cdf_tk.utils import humanize_collection

        suffix = "s are" if len(self.variables) > 1 else " is"
        variables = humanize_collection(self.variables, sort=True, bind_word="and")
        return f"The environment variable{suffix} missing: {variables}"


@dataclass(frozen=True)
class StreamlitRequirementsWarning(FileReadWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.MEDIUM
    dependencies: str | list[str]

    def get_message(self) -> str:
        if isinstance(self.dependencies, str):
            return f"Missing recommended dependencies in requirements.txt: {self.dependencies}"
        return f"Missing dependencies in requirements.txt: {', '.join(self.dependencies)}"


@dataclass(frozen=True)
class ResourceFormatWarning(FileReadWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    errors: tuple[str, ...]

    def get_message(self) -> str:
        sep = "\n     * "
        errors = sep.join(map(escape, self.errors))
        s = "s" if len(self.errors) > 1 else ""
        return f"{type(self).__name__} {len(self.errors)} error{s}:{sep}{errors}"



import time
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any, cast, final

from cognite.client.data_classes import (
    ClientCredentials,
    Function,
    FunctionList,
    FunctionSchedule,
    FunctionSchedulesList,
    FunctionScheduleWrite,
    FunctionScheduleWriteList,
    FunctionWrite,
    FunctionWriteList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    FilesAcl,
    FunctionsAcl,
    SessionsAcl,
)
from cognite.client.data_classes.functions import HANDLER_FILE_NAME
from cognite.client.exceptions import CogniteAuthError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk._parameters import ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.functions import FunctionScheduleID
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.resource_classes import FunctionScheduleYAML, FunctionsYAML
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    calculate_directory_hash,
    calculate_hash,
    calculate_secure_hash,
)
from cognite_toolkit._cdf_tk.utils.cdf import read_auth, try_find_error

from .auth_loaders import GroupAllScopedLoader
from .data_organization_loaders import DataSetsLoader
from .group_scoped_loader import GroupResourceScopedLoader


@final
class FunctionLoader(ResourceLoader[str, FunctionWrite, Function, FunctionWriteList, FunctionList]):
    support_drop = True
    folder_name = "functions"
    filename_pattern = (
        r"^(?:(?!schedule).)*$"  # Matches all yaml files except file names who's stem contain *.schedule.
    )
    resource_cls = Function
    resource_write_cls = FunctionWrite
    list_cls = FunctionList
    list_write_cls = FunctionWriteList
    kind = "Function"
    yaml_cls = FunctionsYAML
    dependencies = frozenset({DataSetsLoader, GroupAllScopedLoader})
    _doc_url = "Functions/operation/postFunctions"
    metadata_value_limit = 512
    support_update = False

    class _MetadataKey:
        function_hash = "cognite-toolkit-hash"
        secret_hash = "cdf-toolkit-secret-hash"

    def __init__(self, client: ToolkitClient, build_path: Path | None, console: Console | None):
        super().__init__(client, build_path, console)
        self.data_set_id_by_external_id: dict[str, int] = {}
        self.function_dir_by_external_id: dict[str, Path] = {}

    @property
    def display_name(self) -> str:
        return "functions"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FunctionWrite] | None, read_only: bool
    ) -> list[Capability] | list[Capability]:
        if not items and items is not None:
            return []

        function_actions = (
            [FunctionsAcl.Action.Read] if read_only else [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        )
        file_actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        return [
            FunctionsAcl(function_actions, FunctionsAcl.Scope.All()),
            FilesAcl(file_actions, FilesAcl.Scope.All()),  # Needed for uploading function artifacts
        ]

    @classmethod
    def get_id(cls, item: Function | FunctionWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("Function must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsLoader, item["dataSetExternalId"]

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        if filepath.parent.name != self.folder_name:
            # Functions configs needs to be in the root function folder.
            # This is to allow arbitrary YAML files inside the function code folder.
            return []
        if self.resource_build_path is None:
            raise ValueError("build_path must be set to compare functions as function code must be compared.")

        raw_list = super().load_resource_file(filepath, environment_variables)
        for item in raw_list:
            item_id = self.get_id(item)
            function_rootdir = Path(self.resource_build_path / item_id)
            self.function_dir_by_external_id[item_id] = function_rootdir
            if "metadata" not in item:
                item["metadata"] = {}
            value = self._create_hash_values(function_rootdir)
            item["metadata"][self._MetadataKey.function_hash] = value
            if "secrets" in item:
                item["metadata"][self._MetadataKey.secret_hash] = calculate_secure_hash(item["secrets"])

        return raw_list

    @classmethod
    def _create_hash_values(cls, function_rootdir: Path) -> str:
        root_hash = calculate_directory_hash(
            function_rootdir, exclude_prefixes={".DS_Store"}, ignore_files={".pyc"}, shorten=True
        )
        hash_value = f"/={root_hash}"
        to_search = [function_rootdir]
        while to_search:
            search_dir = to_search.pop()
            for file in sorted(search_dir.glob("*"), key=lambda x: x.relative_to(function_rootdir).as_posix()):
                if file.is_dir():
                    to_search.append(file)
                    continue
                elif file.is_file() and file.suffix == ".pyc":
                    continue
                elif file.is_file() and file.name == ".DS_Store":
                    continue
                file_hash = calculate_hash(file, shorten=True)
                new_entry = f"{file.relative_to(function_rootdir).as_posix()}={file_hash}"
                if len(hash_value) + len(new_entry) > (cls.metadata_value_limit - 1):
                    break
                hash_value += f";{new_entry}"
        return hash_value

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FunctionWrite:
        item_id = self.get_id(resource)
        if ds_external_id := resource.pop("dataSetExternalId", None):
            self.data_set_id_by_external_id[item_id] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if "fileId" not in resource:
            # The fileID is required for the function to be created, but in the `.create` method
            # we first create that file and then set the fileID.
            resource["fileId"] = "<will_be_generated>"
        return FunctionWrite._load(resource)

    def dump_resource(self, resource: Function, local: dict[str, Any] | None = None) -> dict[str, Any]:
        if resource.status == "Failed":
            dumped = self.dump_id(resource.external_id or resource.name)
            dumped["status"] = "Failed"
            return dumped
        dumped = resource.as_write().dump()
        local = local or {}
        for key in ["cpu", "memory", "runtime"]:
            if key not in local:
                # Server set default values
                dumped.pop(key, None)
            elif isinstance(local.get(key), float) and local[key] < dumped[key]:
                # On Azure and AWS, the server sets the CPU and Memory to the default values if the user
                # pass in lower values. We set this to match the local to avoid triggering a redeploy.
                # Note the user will get a warning about this when the function is created.
                if self.client.config.cloud_provider in ("azure", "aws"):
                    dumped[key] = local[key]
                elif self.client.config.cloud_provider == "gcp" and key == "cpu" and local[key] < 1.0:
                    # GCP does not allow CPU to be set to below 1.0
                    dumped[key] = local[key]
                elif self.client.config.cloud_provider == "gcp" and key == "memory" and local[key] < 1.5:
                    # GCP does not allow Memory to be set to below 1.5
                    dumped[key] = local[key]

        for key in ["indexUrl", "extraIndexUrls"]:
            # Only in write (request) format of the function
            if key in local:
                dumped[key] = local[key]
        if "secrets" in dumped and "secrets" not in local:
            # Secrets are masked in the response.
            dumped.pop("secrets")
        elif "secrets" in dumped and "secrets" in local:
            # Note this will be misleading as the secrets might not be the same as the ones in the API.
            # It will be caught by the hash comparison, but can cause confusion for the user if they
            # compare the dumped file with the API.
            dumped["secrets"] = local["secrets"]

        if file_id := dumped.pop("fileId", None):
            # The fileId is not part of the write object.
            function_zip_file = self.client.files.retrieve(id=file_id)
            if function_zip_file and (data_set_id := function_zip_file.data_set_id):
                dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)

        if dumped.get("functionPath") == HANDLER_FILE_NAME and "functionPath" not in local:
            # Remove the default value of the functionPath
            dumped.pop("functionPath", None)

        return dumped

    def _is_activated(self, action: str) -> bool:
        status = self.client.functions.status()
        if status.status == "activated":
            return True
        if status.status == "requested":
            print(
                HighSeverityWarning(
                    f"Function service activation is in progress, cannot {action} functions."
                ).get_message()
            )
            return False
        else:
            print(
                HighSeverityWarning(
                    "Function service has not been activated, activating now, this may take up to 2 hours..."
                ).get_message()
            )
            self.client.functions.activate()
        return False

    def create(self, items: FunctionWriteList) -> FunctionList:
        created = FunctionList([], cognite_client=self.client)
        if not self._is_activated("create"):
            return created
        if self.resource_build_path is None:
            raise ValueError("build_path must be set to compare functions as function code must be compared.")
        for item in items:
            external_id = item.external_id or item.name
            function_rootdir = self.function_dir_by_external_id[external_id]
            file_id = self.client.functions._zip_and_upload_folder(
                name=item.name,
                folder=str(function_rootdir),
                external_id=external_id,
                data_set_id=self.data_set_id_by_external_id.get(external_id),
            )
            # Wait until the files is available
            t0 = time.perf_counter()
            sleep_time = 1.0  # seconds
            for i in range(6):
                file = self.client.files.retrieve(external_id=external_id)
                if file and file.uploaded:
                    break
                time.sleep(sleep_time)
                sleep_time *= 2
            else:
                elapsed_time = time.perf_counter() - t0
                raise ResourceCreationError(
                    f"Failed to create function {external_id}. CDF API timed out after {elapsed_time:.0f} "
                    "seconds while waiting for the function code to be uploaded. Wait and try again.\nIf the"
                    " problem persists, please contact Cognite support."
                )
            item.file_id = file_id
            created_item = self.client.functions.create(item)
            self._warn_if_cpu_or_memory_changed(created_item, item)
            created.append(created_item)
        return created

    @staticmethod
    def _warn_if_cpu_or_memory_changed(created_item: Function, item: FunctionWrite) -> None:
        is_cpu_increased = (
            isinstance(item.cpu, float) and isinstance(created_item.cpu, float) and item.cpu < created_item.cpu
        )
        is_mem_increased = (
            isinstance(item.memory, float)
            and isinstance(created_item.memory, float)
            and item.memory < created_item.memory
        )
        if is_cpu_increased and is_mem_increased:
            prefix = "CPU and Memory"
            suffix = f"CPU {item.cpu} -> {created_item.cpu}, Memory {item.memory} -> {created_item.memory}"
        elif is_cpu_increased:
            prefix = "CPU"
            suffix = f"{item.cpu} -> {created_item.cpu}"
        elif is_mem_increased:
            prefix = "Memory"
            suffix = f"{item.memory} -> {created_item.memory}"
        else:
            return
        # The server sets the CPU and Memory to the default values, if the user pass in a lower value.
        # This happens on Azure and AWS. Warning the user about this.
        LowSeverityWarning(
            f"Function {prefix} is not configurable. Function {item.external_id!r} set {suffix}"
        ).print_warning()

    def retrieve(self, ids: SequenceNotStr[str]) -> FunctionList:
        if not self._is_activated("retrieve"):
            return FunctionList([])
        return self.client.functions.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        functions = self.retrieve(ids)

        self.client.functions.delete(external_id=ids)
        file_ids = {func.file_id for func in functions if func.file_id}
        self.client.files.delete(id=list(file_ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Function]:
        return iter(self.client.functions)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("dataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))
        # Replaced by the toolkit
        spec.discard(ParameterSpec(("fileId",), frozenset({"int"}), is_required=True, _is_nullable=False))
        return spec


@final
class FunctionScheduleLoader(
    ResourceLoader[
        FunctionScheduleID, FunctionScheduleWrite, FunctionSchedule, FunctionScheduleWriteList, FunctionSchedulesList
    ]
):
    folder_name = "functions"
    filename_pattern = r"^.*schedule.*$"  # Matches all yaml files who's stem contain *.schedule
    resource_cls = FunctionSchedule
    resource_write_cls = FunctionScheduleWrite
    list_cls = FunctionSchedulesList
    list_write_cls = FunctionScheduleWriteList
    kind = "Schedule"
    yaml_cls = FunctionScheduleYAML
    dependencies = frozenset({FunctionLoader, GroupResourceScopedLoader, GroupAllScopedLoader})
    _doc_url = "Function-schedules/operation/postFunctionSchedules"
    parent_resource = frozenset({FunctionLoader})
    support_update = False

    _hash_key = "cdf-auth"
    _description_character_limit = 500

    def __init__(self, client: ToolkitClient, build_path: Path | None, console: Console | None):
        super().__init__(client, build_path, console)
        self.authentication_by_id: dict[FunctionScheduleID, ClientCredentials] = {}

    @property
    def display_name(self) -> str:
        return "function schedules"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FunctionScheduleWrite] | None, read_only: bool
    ) -> list[Capability]:
        if not items and items is not None:
            return []

        function_actions = (
            [FunctionsAcl.Action.Read] if read_only else [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        )
        session_actions = (
            [SessionsAcl.Action.List]
            if read_only
            else [SessionsAcl.Action.List, SessionsAcl.Action.Create, SessionsAcl.Action.Delete]
        )

        return [
            FunctionsAcl(function_actions, FunctionsAcl.Scope.All()),
            SessionsAcl(session_actions, SessionsAcl.Scope.All()),
        ]

    @classmethod
    def dump_id(cls, id: FunctionScheduleID) -> dict[str, Any]:
        return id.dump(camel_case=True)

    @classmethod
    def get_id(cls, item: FunctionScheduleWrite | FunctionSchedule | dict) -> FunctionScheduleID:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"functionExternalId", "name"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return FunctionScheduleID(item["functionExternalId"], item["name"])

        if item.function_external_id is None or item.name is None:
            raise ToolkitRequiredValueError("FunctionSchedule must have functionExternalId and Name set.")
        return FunctionScheduleID(item.function_external_id, item.name)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "functionExternalId" in item:
            yield FunctionLoader, item["functionExternalId"]

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = super().load_resource_file(filepath, environment_variables)
        # We need to the auth hash calculation here, as the output of the load_resource_file
        # is used to compare with the CDF resource.
        for resource in resources:
            identifier = self.get_id(resource)
            credentials = read_auth(
                resource.get("authentication"),
                self.client.config,
                identifier,
                "function schedule",
                console=self.console,
            )
            self.authentication_by_id[identifier] = credentials
            auth_hash = calculate_secure_hash(credentials.dump(camel_case=True), shorten=True)
            extra_str = f" {self._hash_key}: {auth_hash}"
            if "description" not in resource:
                resource["description"] = extra_str[1:]
            elif resource["description"].endswith(extra_str[1:]):
                # The hash is already in the description
                ...
            elif len(resource["description"]) + len(extra_str) < self._description_character_limit:
                resource["description"] += f"{extra_str}"
            else:
                LowSeverityWarning(f"Description is too long for schedule {identifier!r}. Truncating...").print_warning(
                    console=self.console
                )
                truncation = self._description_character_limit - len(extra_str) - 3
                resource["description"] = f"{resource['description'][:truncation]}...{extra_str}"
        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FunctionScheduleWrite:
        if "functionId" in resource:
            identifier = self.get_id(resource)
            LowSeverityWarning(f"FunctionId will be ignored in the schedule {identifier!r}").print_warning(
                console=self.console
            )
            resource.pop("functionId", None)

        return FunctionScheduleWrite._load(resource)

    def dump_resource(self, resource: FunctionSchedule, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if "functionId" in dumped and "functionId" not in local:
            dumped.pop("functionId")
        if "authentication" in local:
            # The authentication is not returned in the response, so we need to add it back.
            dumped["authentication"] = local["authentication"]
        return dumped

    def _resolve_functions_ext_id(self, items: FunctionScheduleWriteList) -> FunctionScheduleWriteList:
        functions = FunctionLoader(self.client, None, None).retrieve(
            list(set([item.function_external_id for item in items]))
        )
        for item in items:
            for func in functions:
                if func.external_id == item.function_external_id:
                    item.function_id = func.id  # type: ignore[assignment]
        return items

    def retrieve(self, ids: SequenceNotStr[FunctionScheduleID]) -> FunctionSchedulesList:
        names_by_function: dict[str, set[str]] = defaultdict(set)
        for id_ in ids:
            names_by_function[id_.function_external_id].add(id_.name)
        functions = FunctionLoader(self.client, None, None).retrieve(list(names_by_function))
        schedules = FunctionSchedulesList([])
        for func in functions:
            func_external_id = cast(str, func.external_id)
            function_schedules = self.client.functions.schedules.list(function_id=func.id, limit=-1)
            for schedule in function_schedules:
                schedule.function_external_id = func_external_id
            schedules.extend(
                [schedule for schedule in function_schedules if schedule.name in names_by_function[func_external_id]]
            )
        return schedules

    def create(self, items: FunctionScheduleWriteList) -> FunctionSchedulesList:
        created_list = FunctionSchedulesList([], cognite_client=self.client)
        for item in items:
            id_ = self.get_id(item)
            if id_ not in self.authentication_by_id:
                raise ToolkitRequiredValueError(f"Authentication is missing for schedule {id_!r}")
            client_credentials = self.authentication_by_id[id_]
            try:
                created = self.client.functions.schedules.create(item, client_credentials=client_credentials)
            except CogniteAuthError as e:
                if hint := try_find_error(client_credentials):
                    raise ResourceCreationError(f"Failed to create Function Schedule {id_}: {hint}") from e
                raise e

            # The PySDK mutates the input object, such that function_id is set and function_external_id is None.
            # If we call .get_id on the returned object, it will raise an error we require the function_external_id
            # to be set.
            created.function_external_id = id_.function_external_id
            created_list.append(created)
        return created_list

    def delete(self, ids: SequenceNotStr[FunctionScheduleID]) -> int:
        schedules = self.retrieve(ids)
        count = 0
        for schedule in schedules:
            if schedule.id:
                self.client.functions.schedules.delete(id=schedule.id)
                count += 1
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[FunctionSchedule]:
        if parent_ids is None:
            yield from self.client.functions.schedules
        else:
            for parent_id in parent_ids:
                if not isinstance(parent_id, str):
                    continue
                yield from self.client.functions.schedules(function_external_id=parent_id)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("authentication",), frozenset({"dict"}), is_required=False, _is_nullable=False))
        spec.add(
            ParameterSpec(("authentication", "clientId"), frozenset({"str"}), is_required=False, _is_nullable=False)
        )
        spec.add(
            ParameterSpec(("authentication", "clientSecret"), frozenset({"str"}), is_required=False, _is_nullable=False)
        )
        return spec

    def sensitive_strings(self, item: FunctionScheduleWrite) -> Iterable[str]:
        id_ = self.get_id(item)
        if id_ in self.authentication_by_id:
            yield self.authentication_by_id[id_].client_secret



from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import ProfileCommand


class TestDumpResource:
    def test_profile_assent_centric(self, toolkit_client: ToolkitClient, monkeypatch) -> None:
        results = ProfileCommand().asset_centric(toolkit_client, verbose=False)

        assert len(results) == 7
        assert {item["Resource"] for item in results} == {
            "Assets",
            "Events",
            "Files",
            "TimeSeries",
            "Sequences",
            "Relationships",
            "Labels",
        }
        total_count = sum(int(item["Count"].replace(",", "")) for item in results)
        assert total_count > 0
        total_metadata_count = 0
        for item in results:
            metadata_count = item.get(ProfileCommand.Columns.MetadataKeyCount, "")
            if "-" in metadata_count or not metadata_count:
                continue
            total_metadata_count += int(metadata_count.replace("-", ""))
        assert total_metadata_count > 0



import sys
from typing import Any, Literal, cast

from cognite.client.data_classes import GroupWrite
from pydantic import ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from .base import ToolkitResource
from .capabilities import Capability

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class GroupYAML(ToolkitResource):
    _cdf_resource = GroupWrite
    name: str
    capabilities: list[Capability] | None = None
    metadata: dict[str, str] | None = None

    @model_validator(mode="wrap")
    @classmethod
    def select_group_type(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if cls is not GroupYAML:
            return handler(data)
        if "sourceId" in data:
            return cast(Self, ExternalGroupYAML.model_validate(data))
        elif "members" in data:
            return cast(Self, CDFGroupYAML.model_validate(data))
        raise ValueError("Missing required field: Either 'sourceId' or 'members'")

    @model_serializer(mode="wrap")
    def serialize_group(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        # Capabilities are serialized as empty dicts [{}, {}, ...]
        # This issue arises because Pydantic's serialization mechanism doesn't automatically
        # handle polymorphic serialization for subclasses of Capability.
        # To address this, we include the below to explicitly calling model dump on the capabilities
        serialized_data = handler(self)
        if self.capabilities:
            serialized_data["capabilities"] = [cap.model_dump(**vars(info)) for cap in self.capabilities]
        return serialized_data


class ExternalGroupYAML(GroupYAML):
    source_id: str


class CDFGroupYAML(GroupYAML):
    members: list[str] | Literal["allUserAccounts"]



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import LocationYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def location_yaml_cases() -> Iterable:
    yaml_cases = [
        # Basic case with only required fields
        {"BasicCase": {"externalId": "loc-001", "name": "Test Location"}},
        # With description and parent
        {
            "ParentCase": {
                "externalId": "loc-002",
                "name": "Child Location",
                "description": "Test description",
                "parentExternalId": "loc-001",
            }
        },
        # With scene
        {
            "SceneCase": {
                "externalId": "loc-003",
                "name": "Location with Scene",
                "scene": {"externalId": "scene-001", "space": "test-space"},
            }
        },
        # With data models
        {
            "DataModelCase": {
                "externalId": "loc-004",
                "name": "Location with Data Models",
                "dataModels": [
                    {"externalId": "model-001", "space": "test-space", "version": "1.0"},
                    {"externalId": "model-002", "space": "test-space", "version": "2.0"},
                ],
            }
        },
        # With instance spaces
        {
            "InstanceSpacesCase": {
                "externalId": "loc-005",
                "name": "Location with Instance Spaces",
                "instanceSpaces": ["space-1", "space-2"],
            }
        },
        # With views
        {
            "ViewsCase": {
                "externalId": "loc-006",
                "name": "Location with Views",
                "views": [
                    {"externalId": "view-001", "space": "test-space", "version": "1.0", "representsEntity": "ASSET"},
                    {
                        "externalId": "view-002",
                        "space": "test-space",
                        "version": "1.0",
                        "representsEntity": "NOTIFICATION",
                    },
                ],
            }
        },
        # With asset centric resources - simple case
        {
            "SimpleAssetCentricCase": {
                "externalId": "loc-007",
                "name": "Location with Asset Centric Resource",
                "assetCentric": {
                    "dataSetExternalIds": ["dataset-001", "dataset-002"],
                    "assetSubtreeExternalIds": [{"externalId": "asset-001"}, {"externalId": "asset-002"}],
                    "externalIdPrefix": "test-prefix",
                },
            }
        },
        # With asset centric resources - full case
        {
            "FullAssetCentricCase": {
                "externalId": "loc-008",
                "name": "Location with Full Asset Centric Resources",
                "assetCentric": {
                    "assets": {
                        "dataSetExternalIds": ["dataset-001"],
                        "assetSubtreeExternalIds": [{"externalId": "asset-001"}],
                        "externalIdPrefix": "asset-prefix",
                    },
                    "events": {
                        "dataSetExternalIds": ["dataset-002"],
                        "assetSubtreeExternalIds": [{"externalId": "asset-002"}],
                        "externalIdPrefix": "event-prefix",
                    },
                    "timeseries": {"dataSetExternalIds": ["dataset-003"], "externalIdPrefix": "ts-prefix"},
                    "files": {
                        "assetSubtreeExternalIds": [{"externalId": "asset-003"}],
                    },
                    "sequences": {"externalIdPrefix": "seq-prefix"},
                },
            }
        },
        # Full example with all fields
        {
            "CompleteCase": {
                "externalId": "loc-009",
                "name": "Full Location Example",
                "description": "Complete location with all fields",
                "parentExternalId": "loc-001",
                "dataModels": [{"externalId": "model-001", "space": "test-space", "version": "1.0"}],
                "instanceSpaces": ["space-1", "space-2"],
                "scene": {"externalId": "scene-001", "space": "test-space"},
                "views": [
                    {"externalId": "view-001", "space": "test-space", "version": "1.0", "representsEntity": "ASSET"}
                ],
                "assetCentric": {
                    "assets": {"externalIdPrefix": "asset-prefix"},
                    "events": {"externalIdPrefix": "event-prefix"},
                },
            }
        },
    ]

    yield from (pytest.param(next(iter(case.values())), id=next(iter(case.keys()))) for case in yaml_cases)


def invalid_location_filters_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Location 1"}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    yield pytest.param(
        {"externalId": "my_location", "name": "Location 1", "dataModelingType": "ASSET_CENTRIC_ONLY"},
        {"In field dataModelingType input should be 'HYBRID' or 'DATA_MODELING_ONLY'. Got 'ASSET_CENTRIC_ONLY'."},
        id="Invalid dataModelingType value",
    )
    yield pytest.param(
        {"externalId": "location_1", "name": "Location 1", "dataModels": ["model-1", "model-2"]},
        {
            "In dataModels input must be an object of type DataModelID. Got 'model-1' of type str.",
            "In dataModels input must be an object of type DataModelID. Got 'model-2' of type str.",
        },
        id="Invalid list of dataModels.",
    )


class TestLocationYAML:
    @pytest.mark.parametrize("data", list(find_resources("Location")))
    def test_load_valid_location(self, data: dict[str, object]) -> None:
        loaded = LocationYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("test_input", location_yaml_cases())
    def test_valid_location_variants(self, test_input: dict) -> None:
        location = LocationYAML.model_validate(test_input)
        assert location.model_dump(by_alias=True, exclude_unset=True) == test_input

    @pytest.mark.parametrize("data, expected_errors", list(invalid_location_filters_test_cases()))
    def test_invalid_location_filters_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, LocationYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from collections.abc import Iterable
from pathlib import Path

import pytest
import yaml
from cognite.client.data_classes import TimeSeries

from cognite_toolkit._cdf_tk._parameters import ParameterSpecSet
from cognite_toolkit._cdf_tk.loaders import ContainerLoader, SpaceLoader, TimeSeriesLoader, ViewLoader
from cognite_toolkit._cdf_tk.tk_warnings import (
    CaseTypoWarning,
    DataSetMissingWarning,
    MissingRequiredParameterWarning,
    ToolkitWarning,
    UnusedParameterWarning,
)
from cognite_toolkit._cdf_tk.validation import validate_data_set_is_set, validate_resource_yaml
from tests.data import LOAD_DATA

DUMMY_FILE = Path("dummy.yaml")


def test_validate_raw() -> None:
    raw_file = LOAD_DATA / "timeseries" / "wrong_case.yaml"

    warnings = validate_resource_yaml(
        yaml.safe_load(raw_file.read_text()), TimeSeriesLoader.get_write_cls_parameter_spec(), raw_file
    )

    assert len(warnings) == 2
    assert sorted(warnings) == sorted(
        [
            CaseTypoWarning(raw_file, 1, ("is_string",), "is_string", "isString"),
            CaseTypoWarning(raw_file, 1, ("is_step",), "is_step", "isStep"),
        ]
    )


def test_validate_raw_nested() -> None:
    raw_file = LOAD_DATA / "datamodels" / "snake_cased_view_property.yaml"
    warnings = validate_resource_yaml(
        yaml.safe_load(raw_file.read_text()), ViewLoader.get_write_cls_parameter_spec(), raw_file
    )

    assert len(warnings) == 1
    assert warnings == [
        CaseTypoWarning(
            raw_file,
            None,
            ("properties", "criticality", "container_property_identifier"),
            "container_property_identifier",
            "containerPropertyIdentifier",
        )
    ]


def test_validate_data_set_is_set():
    warnings = validate_data_set_is_set(
        {"externalId": "myTimeSeries", "name": "My Time Series"}, TimeSeries, Path("timeseries.yaml")
    )

    assert sorted(warnings) == sorted(
        [DataSetMissingWarning(Path("timeseries.yaml"), "myTimeSeries", "externalId", "TimeSeries")]
    )


def validate_yaml_config_test_cases() -> Iterable:
    content = """space: my_space
"""
    yield pytest.param(content, SpaceLoader.get_write_cls_parameter_spec(), [], id="Valid space")

    content = """space: my_space
nme: My space
"""
    yield pytest.param(
        content,
        SpaceLoader.get_write_cls_parameter_spec(),
        [
            UnusedParameterWarning(
                DUMMY_FILE,
                None,
                ("nme",),
                "nme",
            )
        ],
        id="Unused parameter",
    )
    content = """externalId: Pump
name: Pump
usedFor: node
properties:
  DesignPointFlowGPM:
    autoIncrement: false
    default_value: astring
    nullable: true
    type:
      list: false
      type: float64"""
    yield pytest.param(
        content,
        ContainerLoader.get_write_cls_parameter_spec(),
        [
            MissingRequiredParameterWarning(DUMMY_FILE, None, ("space",), "space"),
            CaseTypoWarning(
                DUMMY_FILE, None, ("properties", "DesignPointFlowGPM", "default_value"), "default_value", "defaultValue"
            ),
        ],
        id="Missing required and misspelling",
    )
    content = """
- space: sp_my_space
  name: First space
- space: sp_my_space_2
  nme: Second space
- space: sp_my_space_3
  nme: Third space"""
    yield pytest.param(
        content,
        SpaceLoader.get_write_cls_parameter_spec(),
        [
            UnusedParameterWarning(DUMMY_FILE, 2, ("nme",), "nme"),
            UnusedParameterWarning(DUMMY_FILE, 3, ("nme",), "nme"),
        ],
        id="Unused parameter in list",
    )


class TestValidateYAML:
    @pytest.mark.parametrize("content, spec, expected_warnings", list(validate_yaml_config_test_cases()))
    def test_validate_yaml_config(self, content: str, spec: ParameterSpecSet, expected_warnings: list[ToolkitWarning]):
        data = yaml.safe_load(content)
        warnings = validate_resource_yaml(data, spec, DUMMY_FILE)
        assert sorted(warnings) == sorted(expected_warnings)



from __future__ import annotations

import shutil
from collections import defaultdict
from collections.abc import Collection, Iterator, Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, SupportsIndex, overload

from cognite_toolkit._cdf_tk.constants import INDEX_PATTERN
from cognite_toolkit._cdf_tk.utils import calculate_directory_hash, iterate_modules, resource_folder_from_path

from ._module_toml import ModuleToml


@dataclass(frozen=True)
class ModuleLocation:
    """This represents the location of a module in a directory structure.
    Args:
        dir: The absolute path to the module directory.
        source_absolute_path: The absolute path to the source directory.
        is_selected: Whether the module is selected by the user.
        source_paths: The paths to all files in the module.
    """

    dir: Path
    source_absolute_path: Path
    source_paths: list[Path]
    is_selected: bool = False
    definition: ModuleToml | None = None

    @property
    def has_example_data(self) -> bool:
        return bool(self.definition and self.definition.data)

    @property
    def name(self) -> str:
        """The name of the module."""
        return self.dir.name

    @property
    def title(self) -> str | None:
        """The title of the module."""
        if self.definition:
            return self.definition.title
        return None

    @property
    def relative_path(self) -> Path:
        """The relative path to the module."""
        return self.dir.relative_to(self.source_absolute_path)

    @property
    def module_selections(self) -> set[str | Path]:
        """Ways of selecting this module."""
        return {self.name, self.relative_path, *self.parent_relative_paths}

    @property
    def is_selected_by_default(self) -> bool:
        """Whether the module is default selected."""
        if self.definition:
            return self.definition.is_selected_by_default
        return False

    @cached_property
    def parent_relative_paths(self) -> set[Path]:
        """All relative parent paths of the module."""
        return set(self.relative_path.parents)

    @cached_property
    def hash(self) -> str:
        """The hash of the module."""
        return calculate_directory_hash(self.dir, shorten=True)

    @cached_property
    def resource_directories(self) -> set[str]:
        """The resource directories in the module."""
        source_path_by_resource_folder, _ = self._source_paths_by_resource_folder
        return set(source_path_by_resource_folder.keys())

    @property
    def _source_paths_by_resource_folder(self) -> tuple[dict[str, list[Path]], set[str]]:
        """The source paths grouped by resource folder."""
        source_paths_by_resource_folder = defaultdict(list)
        # The directories in the module that are not resource directories.
        invalid_resource_directory: set[str] = set()
        for filepath in self.source_paths:
            try:
                resource_folder = resource_folder_from_path(filepath)
            except ValueError:
                relative_to_module = filepath.relative_to(self.dir)
                is_file_in_resource_folder = relative_to_module.parts[0] == filepath.name
                if not is_file_in_resource_folder:
                    invalid_resource_directory.add(relative_to_module.parts[0])
                continue
            if filepath.is_file():
                source_paths_by_resource_folder[resource_folder].append(filepath)
        return source_paths_by_resource_folder, invalid_resource_directory

    @cached_property
    def source_paths_by_resource_folder(self) -> dict[str, list[Path]]:
        """The source paths grouped by resource folder."""
        source_paths_by_resource_folder, _ = self._source_paths_by_resource_folder

        # Sort to support 1., 2. etc prefixes
        def sort_key(p: Path) -> tuple[int, int, str]:
            first = {
                ".yaml": 0,
                ".yml": 0,
            }.get(p.suffix.lower(), 1)
            # We ensure that the YAML files are sorted before other files.
            # This is when we add indexes to files. We want to ensure that, for example, a .sql file
            # with the same name as a .yaml file gets the same index as the .yaml file.
            if result := INDEX_PATTERN.search(p.stem):
                return first, int(result.group()[:-1]), p.name
            else:
                return first, len(filepaths) + 1, p.name

        for filepaths in source_paths_by_resource_folder.values():
            # The builder of a module can control the order that resources are deployed by prefixing a number
            # The custom key 'sort_key' is to get the sort on integer and not the string.
            filepaths.sort(key=sort_key)

        return source_paths_by_resource_folder

    @cached_property
    def not_resource_directories(self) -> set[str]:
        """The directories in the module that are not resource directories."""
        return self._source_paths_by_resource_folder[1]

    @cached_property
    def dependencies(self) -> set[str]:
        """The dependencies of the module."""
        if self.definition:
            return set(self.definition.dependencies)
        return set()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, is_selected={self.is_selected}, file_count={len(self.source_paths)})"

    def __str__(self) -> str:
        return self.name

    def as_read_module(self) -> ReadModule:
        return ReadModule(
            dir=self.dir,
            resource_directories=tuple(self.resource_directories),
        )


@dataclass(frozen=True)
class ReadModule:
    """This is a short representation of a module.

    Args:
        dir: The absolute path to the module directory.
        resource_directories: The resource directories in the module.
    """

    dir: Path
    resource_directories: tuple[str, ...]

    def resource_dir_path(self, resource_folder: str) -> Path | None:
        """Returns the path to a resource in the module.

        Args:
            resource_folder: The name of the resource.

        Returns:
            The path to the resource if it exists, otherwise None.
        """
        for resource_dir in self.resource_directories:
            if resource_dir == resource_folder and (resource_path := self.dir / resource_folder).exists():
                return resource_path
        return None

    @classmethod
    def load(cls, data: dict[str, Any]) -> ReadModule:
        return cls(
            dir=Path(data["dir"]),
            resource_directories=tuple(data["resource_directories"]),
        )

    def dump(self) -> dict[str, Any]:
        return {
            "dir": self.dir.as_posix(),
            "resource_directories": list(self.resource_directories),
        }


class ModuleDirectories(tuple, Sequence[ModuleLocation]):
    """This is an internal representation of the module directories in a source directory.

    The motivation for this class is to provide helper functions for the user to interact with the module directories.
    """

    # Subclassing tuple to make the class immutable. ModuleDirectories is expected to be initialized and
    # then used as a read-only object.
    def __new__(cls, collection: Collection[ModuleLocation] | None) -> ModuleDirectories:
        # Need to override __new__ to as we are subclassing a tuple:
        #   https://stackoverflow.com/questions/1565374/subclassing-tuple-with-multiple-init-arguments
        return super().__new__(cls, tuple(collection or []))

    def __init__(self, collection: Collection[ModuleLocation] | None) -> None: ...

    @cached_property
    def available(self) -> set[str | Path]:
        return {selection for module_location in self for selection in module_location.module_selections}

    @cached_property
    def selected(self) -> ModuleDirectories:
        return ModuleDirectories([module for module in self if module.is_selected])

    @cached_property
    def available_paths(self) -> set[Path]:
        return {item for item in self.available if isinstance(item, Path)}

    @cached_property
    def available_names(self) -> set[str]:
        return {item for item in self.available if isinstance(item, str)}

    @classmethod
    def load(
        cls,
        organization_dir: Path,
        user_selected_modules: set[str | Path] | None = None,
    ) -> ModuleDirectories:
        """Loads the modules in the source directory.

        Args:
            organization_dir: The absolute path to the source directory.
            user_selected_modules: The modules selected by the user either by name or by path.

        """
        # Assume all modules are selected if no selection is given.
        user_selected_modules = user_selected_modules or {Path("")}

        module_locations: list[ModuleLocation] = []
        for module, source_paths in iterate_modules(organization_dir):
            relative_module_dir = module.relative_to(organization_dir)
            module_toml: ModuleToml | None = None
            tags: set[str] = set()
            if (module / ModuleToml.filename).exists():
                module_toml = ModuleToml.load(module / ModuleToml.filename)
                tags = set(module_toml.tags)

            module_locations.append(
                ModuleLocation(
                    module,
                    organization_dir,
                    source_paths,
                    cls._is_selected_module(relative_module_dir, user_selected_modules, tags),
                    module_toml,
                )
            )

        return cls(module_locations)

    def dump(self, organization_dir: Path) -> None:
        """Dumps the module directories to the source directory.

        Args:
            organization_dir: The absolute path to the source directory.
        """
        for module in self:
            module_dir = organization_dir / module.relative_path
            module_dir.mkdir(parents=True, exist_ok=True)
            for source_file in module.source_paths:
                relative_file_path = source_file.relative_to(module.dir)
                absolute_file_path = module_dir / relative_file_path
                absolute_file_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(source_file, absolute_file_path)

    @classmethod
    def _is_selected_module(
        cls, relative_module_dir: Path, user_selected: set[str | Path], module_tags: set[str]
    ) -> bool:
        """Checks whether a module is selected by the user."""
        return (
            relative_module_dir.name in user_selected
            or relative_module_dir in user_selected
            or any(parent in user_selected for parent in relative_module_dir.parents)
            # Check if the module has any tags that the user has selected,
            # i.e., that the intersection of the module tags and the user selected tags is not empty.
            or bool(module_tags & user_selected)
        )

    def as_path_by_name(self) -> dict[str, list[Path]]:
        module_path_by_name: dict[str, list[Path]] = defaultdict(list)
        for module in self:
            module_path_by_name[module.name].append(module.relative_path)
        return module_path_by_name

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[ModuleLocation]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> ModuleLocation: ...

    @overload
    def __getitem__(self, index: slice) -> ModuleDirectories: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> ModuleLocation | ModuleDirectories:
        if isinstance(index, slice):
            return ModuleDirectories(super().__getitem__(index))
        return super().__getitem__(index)



#!/usr/bin/env python
import argparse
import os
import shutil
from pathlib import Path

THIS_FOLDER = Path(__file__).parent.absolute()
DEMO_PROJECT = THIS_FOLDER.parent / "demo_project"

parser = argparse.ArgumentParser()
parser.add_argument("--modules", help="Which modules to run.", type=str, choices=["demo", "all"], default="demo")


def run() -> None:
    args = parser.parse_args()
    print("Running copy commands to prep deployment of demo...")
    os.makedirs(DEMO_PROJECT, exist_ok=True)
    print("Copying my demo.config.yaml to root of repo...")

    demo_config_path = THIS_FOLDER / "config.demo.yaml"
    destination_path = DEMO_PROJECT / "config.demo.yaml"
    # By default the 'selected_modules_and_packages' is the selected demo packages/modules
    if args.modules == "all":
        import yaml

        demo_config = yaml.safe_load(demo_config_path.read_text())
        demo_config["environment"]["selected"] = ["modules/"]
        destination_path.write_text(yaml.dump(demo_config))
        print(f"Updated {demo_config_path} to include all modules.")
    else:
        shutil.copy(demo_config_path, destination_path)
        print(f"Copied {demo_config_path} to {destination_path} with selected demo packages.")


if __name__ == "__main__":
    run()



import shutil
from collections.abc import Callable, Iterable, Sequence

from cognite_toolkit._cdf_tk.builders import Builder
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    ModuleLocation,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileExistsError, ToolkitNotADirectoryError, ToolkitValueError
from cognite_toolkit._cdf_tk.loaders import StreamlitLoader
from cognite_toolkit._cdf_tk.tk_warnings import (
    FileReadWarning,
    HighSeverityWarning,
    StreamlitRequirementsWarning,
    ToolkitWarning,
    WarningList,
)
from cognite_toolkit._cdf_tk.utils.file import safe_read


class StreamlitBuilder(Builder):
    _resource_folder = StreamlitLoader.folder_name

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | Sequence[ToolkitWarning]]:
        for source_file in source_files:
            if source_file.loaded is None:
                continue
            if source_file.source.path.parent.parent != module.dir:
                # Streamlit YAML files must be in the resource folder top level
                continue

            loader, warning = self._get_loader(source_file.source.path)
            if loader is None:
                if warning is not None:
                    yield [warning]
                continue

            warnings = WarningList[FileReadWarning]()
            if loader is StreamlitLoader:
                warnings = self.copy_app_directory_to_build(source_file)

            destination_path = self._create_destination_path(source_file.source.path, loader.kind)

            yield BuildDestinationFile(
                path=destination_path,
                loaded=source_file.loaded,
                loader=loader,
                source=source_file.source,
                extra_sources=None,
                warnings=warnings,
            )

    def copy_app_directory_to_build(self, source_file: BuildSourceFile) -> WarningList[FileReadWarning]:
        raw_content = source_file.loaded
        if raw_content is None:
            # This should already be checked before calling this method.
            raise ToolkitValueError("Streamlit source file should be a YAML file.")
        raw_apps = raw_content if isinstance(raw_content, list) else [raw_content]
        warnings = WarningList[FileReadWarning]()
        for raw_app in raw_apps:
            external_id = raw_app.get("externalId")
            if not external_id:
                warnings.append(
                    HighSeverityWarning(
                        f"StreamlitApp in {source_file.source.path.as_posix()!r} has no externalId defined. "
                        f"This is used to match the Streamlit App to directory."
                    )
                )
                continue
            app_directory = source_file.source.path.with_name(external_id)

            if not app_directory.is_dir():
                raise ToolkitNotADirectoryError(
                    f"StreamlitApp directory not found in {app_directory}(based on externalId {external_id} defined in {source_file.source.path.as_posix()!r}.)"
                )

            requirements_file_content = safe_read(app_directory / "requirements.txt").splitlines()
            missing_packages = StreamlitLoader._missing_recommended_requirements(requirements_file_content)
            if len(missing_packages) > 0:
                warnings.append(StreamlitRequirementsWarning(app_directory / "requirements.txt", missing_packages))

            destination = self.build_dir / self.resource_folder / external_id
            if destination.exists():
                raise ToolkitFileExistsError(
                    f"StreamlitApp {external_id!r} is duplicated. If this is unexpected, ensure you have a clean build directory."
                )
            shutil.copytree(app_directory, destination, ignore=shutil.ignore_patterns("__pycache__"))

        return warnings



from __future__ import annotations

import re
import sys
from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar

from rich import print

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.constants import clean_name
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
    ToolkitTOMLFormatError,
    ToolkitVersionError,
)
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning

if sys.version_info >= (3, 11):
    import tomllib
    from tomllib import TOMLDecodeError
else:
    import tomli as tomllib
    from tomli import TOMLDecodeError


@dataclass
class CLIConfig:
    default_organization_dir: Path
    default_env: str = "dev"
    file_encoding: str | None = None
    has_user_set_default_org: bool = False
    has_user_set_default_env: bool = False

    @classmethod
    def load(cls, raw: dict[str, Any], cwd: Path) -> CLIConfig:
        has_user_set_default_org = "default_organization_dir" in raw
        has_user_set_default_env = "default_env" in raw
        default_organization_dir = cwd / raw["default_organization_dir"] if has_user_set_default_org else Path.cwd()
        return cls(
            default_organization_dir=default_organization_dir,
            default_env=raw.get("default_env", "dev"),
            file_encoding=raw.get("file_encoding"),
            has_user_set_default_org=has_user_set_default_org,
            has_user_set_default_env=has_user_set_default_env,
        )


@dataclass
class ModulesConfig:
    version: str
    packages: dict[str, list[str]] = field(default_factory=dict)

    @classmethod
    def load(cls, raw: dict[str, Any]) -> ModulesConfig:
        version = raw["version"]
        packages = raw.get("packages", {})
        if (
            version != _version.__version__
            and (len(sys.argv) > 2 and sys.argv[1:3] != ["modules", "upgrade"])
            and "--help" not in sys.argv
        ):
            raise ToolkitVersionError(
                f"The version of the modules ({version}) does not match the version of the installed CLI "
                f"({_version.__version__}). Please either run `cdf-tk modules upgrade` to upgrade the modules OR "
                f"run `pip install cognite-toolkit=={version}` to downgrade cdf-tk CLI."
            )
        return cls(version=version, packages=packages)


@dataclass
class CDFToml:
    """This is the configuration for the CLI and Modules"""

    file_name: ClassVar[str] = "cdf.toml"

    cdf: CLIConfig
    modules: ModulesConfig
    alpha_flags: dict[str, bool] = field(default_factory=dict)
    plugins: dict[str, bool] = field(default_factory=dict)

    is_loaded_from_file: bool = False

    @classmethod
    def load(cls, cwd: Path | None = None, use_singleton: bool = True) -> CDFToml:
        """Loads the cdf.toml file from the given path. If use_singleton is True, the instance will be stored as a
        singleton and returned on subsequent calls."""
        global _CDF_TOML
        if use_singleton and _CDF_TOML:
            return _CDF_TOML
        cwd = cwd or Path.cwd()
        file_path = cwd / cls.file_name
        if file_path.exists():
            raw = _read_toml(file_path)

            # No required fields in the cdf section
            cdf = CLIConfig.load(raw["cdf"], cwd) if "cdf" in raw else CLIConfig(cwd)
            try:
                modules = ModulesConfig.load(raw["modules"])
            except KeyError as e:
                raise ToolkitRequiredValueError(f"Missing required value in {cls.file_name}: {e.args}")
            alpha_flags = {}
            if "alpha_flags" in raw:
                alpha_flags = {clean_name(k): v for k, v in raw["alpha_flags"].items()}
            if not alpha_flags and "feature_flags" in raw:
                MediumSeverityWarning(
                    "The 'feature_flags' section has been renamed to 'alpha_flags'. Please update your cdf.toml file."
                ).print_warning()
                alpha_flags = {clean_name(k): v for k, v in raw["feature_flags"].items()}

            plugins = {}
            if "plugins" in raw:
                plugins = {clean_name(k): v for k, v in raw["plugins"].items()}

            instance = cls(cdf=cdf, modules=modules, alpha_flags=alpha_flags, plugins=plugins, is_loaded_from_file=True)
            if use_singleton:
                _CDF_TOML = instance
            return instance
        else:
            return cls(
                cdf=CLIConfig(cwd),
                modules=ModulesConfig.load({"version": _version.__version__}),
                alpha_flags={},
                plugins={},
                is_loaded_from_file=False,
            )


def _read_toml(file_path: Path) -> dict[str, Any]:
    # TOML files are required to be UTF-8 encoded
    content = file_path.read_text(encoding="utf-8")
    try:
        return tomllib.loads(content)
    except TOMLDecodeError as e:
        if file_path.is_relative_to(Path.cwd()):
            file_path = file_path.relative_to(Path.cwd())
        extra = ""
        with suppress(Exception):
            # If any errors is raised here, we ignore it and continue
            if result := re.search(r"\(at line (\d+), column (\d+)\)", str(e)):
                line_no = int(result.group(1))
                column_location = int(result.group(2))
                lines = content.splitlines()
                line = lines[line_no - 1]
                prefix = "\n    [blue]HINT: [/blue]"
                if column_location >= len(line):
                    extra = f"{prefix}This is near the end of the line '{line}'"
                elif column_location == 0:
                    extra = f"{prefix}This is near the beginning of the line '{line}'"
                else:
                    extra = f"{prefix}This is near the '{line[:column_location]}' in '{line[column_location:]}'"

        err = ToolkitTOMLFormatError(f"Error reading {file_path.as_posix()!r}: {e!s}{extra}")
        print(f"  [bold red]ERROR ([/][red]{type(err).__name__}[/][bold red]):[/] {err}")
        # We read the CDF.TML file from the module level, so the app has not been loaded yet.
        # Therefore, we raise SystemExit to stop the execution
        raise SystemExit(1)


_CDF_TOML: CDFToml | None = None

if __name__ == "__main__":
    from pprint import pprint

    # This is a test to quickly check that the code works.
    # also useful to check that when you change cdf.toml it is loaded correctly
    _ROOT = Path(__file__).parent.parent.parent
    pprint(CDFToml.load(_ROOT))  # noqa: T203



import contextlib
import os
from collections.abc import Iterator
from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent


@contextlib.contextmanager
def chdir(new_dir: Path) -> Iterator[None]:
    """
    Change directory to new_dir and return to the original directory when exiting the context.

    Args:
        new_dir: The new directory to change to.

    """
    current_working_dir = Path.cwd()
    os.chdir(new_dir)

    try:
        yield

    finally:
        os.chdir(current_working_dir)
